from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING

import click
from alembic.config import Config

from ai.backend.common.types import AgentId
from ai.backend.logging import BraceStyleAdapter
from ai.backend.logging.utils import enforce_debug_logging
from ai.backend.manager.errors.resource import ConfigurationLoadFailed

if TYPE_CHECKING:
    from .context import CLIContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@click.group()
def cli() -> None:
    pass


@cli.command()
@click.argument("agent_id", type=str)
@click.option(
    "-f",
    "--alembic-config",
    default="alembic.ini",
    metavar="PATH",
    type=click.Path(exists=True, dir_okay=False),
    help="The path to Alembic config file. [default: alembic.ini]",
)
@click.option(
    "-t",
    "--timeout",
    default=10.0,
    type=float,
    help="The timeout to wait until declaring failure. [default: 10.0]",
)
@click.pass_obj
def ping(cli_ctx: CLIContext, agent_id: str, alembic_config: str, timeout: float) -> None:
    """
    Ping the agent with AGENT_ID to check whether it responds to an RPC call.

    It uses AgentRPCCache to make the actual RPC call, which reads the agent address and public key
    from the PostgreSQL database.  If the target agent have changed its address or public key while
    the manager is *not* running, it may fail even when the agent is alive.

    This command temporarily enables the DEBUG-level logging for ai.backend.manager.agent_cache
    and callosum to help debugging for when there are connection issues, regardless of the logging
    configuration in manager.toml.
    """
    from zmq.auth.certs import load_certificate

    from ai.backend.common.auth import PublicKey, SecretKey
    from ai.backend.manager.agent_cache import AgentRPCCache
    from ai.backend.manager.models.utils import create_async_engine

    async def _impl() -> None:
        bootstrap_config = await cli_ctx.get_bootstrap_config()
        manager_public_key, manager_secret_key = load_certificate(
            bootstrap_config.manager.rpc_auth_manager_keypair
        )
        if manager_secret_key is None:
            raise ConfigurationLoadFailed("Manager secret key is not available in the keypair")
        alembic_cfg = Config(alembic_config)
        sa_url = alembic_cfg.get_main_option("sqlalchemy.url")
        db = create_async_engine(sa_url)
        agent_cache = AgentRPCCache(
            db,
            manager_public_key=PublicKey(manager_public_key),
            manager_secret_key=SecretKey(manager_secret_key),
        )
        try:
            log.info("Contacting ag:{} ...", agent_id)
            enforce_debug_logging(["callosum", "ai.backend.manager.agent_cache"])
            async with agent_cache.rpc_context(
                AgentId(agent_id),
                invoke_timeout=timeout,
            ) as rpc:
                result = await rpc.call.ping("pong")
                print(f"Received response from ag:{agent_id}: {result}")
                # FIXME: Use gather_hwinfo() when we implement get_node_hwinfo() of CPUPlugin and MemoryPlugin like below.
                # result = await rpc.call.gather_hwinfo()
                # print(f"Retrieved ag:{agent_id} hardware information as a health check:")
                # pprint(result)
        except TimeoutError:
            log.error("Timeout occurred while reading the response from ag:{}", agent_id)
        except Exception:
            log.exception("Exception occurred while reading the response from ag:{}", agent_id)
        finally:
            await db.dispose()

    asyncio.run(_impl())



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, override
from uuid import UUID

from ai.backend.common.data.notification import (
    EmailSpec,
    NotificationChannelType,
    NotificationRuleType,
    WebhookSpec,
)
from ai.backend.manager.types import OptionalState, PartialModifier


@dataclass
class NotificationChannelData:
    """Domain model data for notification channel."""

    id: UUID
    name: str
    description: str | None
    channel_type: NotificationChannelType
    spec: WebhookSpec | EmailSpec
    enabled: bool
    created_by: UUID
    created_at: datetime = field(compare=False)
    updated_at: datetime = field(compare=False)


@dataclass
class NotificationRuleData:
    """Domain model data for notification rule."""

    id: UUID
    name: str
    description: str | None
    rule_type: NotificationRuleType
    channel: NotificationChannelData
    message_template: str
    enabled: bool
    created_by: UUID
    created_at: datetime = field(compare=False)
    updated_at: datetime = field(compare=False)


@dataclass
class NotificationChannelModifier(PartialModifier):
    """Modifier for notification channel."""

    name: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    description: OptionalState[str | None] = field(default_factory=OptionalState[str | None].nop)
    spec: OptionalState[WebhookSpec | EmailSpec] = field(
        default_factory=OptionalState[WebhookSpec | EmailSpec].nop
    )
    enabled: OptionalState[bool] = field(default_factory=OptionalState[bool].nop)

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.description.update_dict(to_update, "description")
        spec_value = self.spec.optional_value()
        if spec_value is not None:
            to_update["spec"] = spec_value.model_dump()
        self.enabled.update_dict(to_update, "enabled")
        return to_update


@dataclass
class NotificationRuleModifier(PartialModifier):
    """Modifier for notification rule."""

    name: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    description: OptionalState[str | None] = field(default_factory=OptionalState[str | None].nop)
    message_template: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    enabled: OptionalState[bool] = field(default_factory=OptionalState[bool].nop)

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.description.update_dict(to_update, "description")
        self.message_template.update_dict(to_update, "message_template")
        self.enabled.update_dict(to_update, "enabled")
        return to_update


@dataclass
class NotificationChannelListResult:
    """Search result with total count for notification channels."""

    items: list[NotificationChannelData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


@dataclass
class NotificationRuleListResult:
    """Search result with total count for notification rules."""

    items: list[NotificationRuleData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool



"""Service layer for app configuration operations."""

from __future__ import annotations

import logging

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.models.app_config import AppConfigScopeType
from ai.backend.manager.repositories.app_config import AppConfigRepository

from .actions import (
    DeleteDomainConfigAction,
    DeleteDomainConfigActionResult,
    DeleteUserConfigAction,
    DeleteUserConfigActionResult,
    GetDomainConfigAction,
    GetDomainConfigActionResult,
    GetMergedAppConfigAction,
    GetMergedAppConfigActionResult,
    GetUserConfigAction,
    GetUserConfigActionResult,
    UpsertDomainConfigAction,
    UpsertDomainConfigActionResult,
    UpsertUserConfigAction,
    UpsertUserConfigActionResult,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AppConfigService:
    """Service for app configuration operations."""

    _app_config_repository: AppConfigRepository

    def __init__(
        self,
        app_config_repository: AppConfigRepository,
    ) -> None:
        self._app_config_repository = app_config_repository

    # Domain config operations

    async def get_domain_config(self, action: GetDomainConfigAction) -> GetDomainConfigActionResult:
        """Get domain-level app configuration."""
        log.debug("Getting domain config for: {}", action.domain_name)
        config_data = await self._app_config_repository.get_config(
            AppConfigScopeType.DOMAIN,
            action.domain_name,
        )
        return GetDomainConfigActionResult(result=config_data)

    async def upsert_domain_config(
        self, action: UpsertDomainConfigAction
    ) -> UpsertDomainConfigActionResult:
        """Create or update domain-level app configuration."""
        log.debug("Upserting domain config for: {}", action.domain_name)
        config_data = await self._app_config_repository.upsert_config(
            AppConfigScopeType.DOMAIN,
            action.domain_name,
            action.updater_spec,
        )
        return UpsertDomainConfigActionResult(result=config_data)

    async def delete_domain_config(
        self, action: DeleteDomainConfigAction
    ) -> DeleteDomainConfigActionResult:
        """Delete domain-level app configuration."""
        log.debug("Deleting domain config for: {}", action.domain_name)
        deleted = await self._app_config_repository.delete_config(
            AppConfigScopeType.DOMAIN,
            action.domain_name,
        )
        return DeleteDomainConfigActionResult(
            deleted=deleted,
            domain_name=action.domain_name,
        )

    # User config operations

    async def get_user_config(self, action: GetUserConfigAction) -> GetUserConfigActionResult:
        """Get user-level app configuration."""
        log.debug("Getting user config for: {}", action.user_id)
        config_data = await self._app_config_repository.get_config(
            AppConfigScopeType.USER,
            action.user_id,
        )
        return GetUserConfigActionResult(result=config_data)

    async def upsert_user_config(
        self, action: UpsertUserConfigAction
    ) -> UpsertUserConfigActionResult:
        """Create or update user-level app configuration."""
        log.debug("Upserting user config for: {}", action.user_id)
        config_data = await self._app_config_repository.upsert_config(
            AppConfigScopeType.USER,
            action.user_id,
            action.updater_spec,
        )
        return UpsertUserConfigActionResult(result=config_data)

    async def delete_user_config(
        self, action: DeleteUserConfigAction
    ) -> DeleteUserConfigActionResult:
        """Delete user-level app configuration."""
        log.debug("Deleting user config for: {}", action.user_id)
        deleted = await self._app_config_repository.delete_config(
            AppConfigScopeType.USER,
            action.user_id,
        )
        return DeleteUserConfigActionResult(
            deleted=deleted,
            user_id=action.user_id,
        )

    # Merged config operation

    async def get_merged_config(
        self, action: GetMergedAppConfigAction
    ) -> GetMergedAppConfigActionResult:
        """
        Get merged app configuration for a user.
        Domain config is merged with user config (user overrides domain).
        """
        log.debug("Getting merged app config for user: {}", action.user_id)
        merged_config = await self._app_config_repository.get_merged_config(action.user_id)
        return GetMergedAppConfigActionResult(
            user_id=action.user_id,
            merged_config=merged_config,
        )



"""Access token resolver functions."""

from __future__ import annotations

import strawberry
from strawberry import Info

from ai.backend.manager.api.gql.deployment.types.access_token import (
    AccessToken,
    CreateAccessTokenInput,
    CreateAccessTokenPayload,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.services.deployment.actions.access_token.create_access_token import (
    CreateAccessTokenAction,
)

# Mutation resolvers


@strawberry.mutation(description="Added in 25.16.0")  # type: ignore[misc]
async def create_access_token(
    input: CreateAccessTokenInput, info: Info[StrawberryGQLContext]
) -> CreateAccessTokenPayload:
    """Create a new access token for a deployment."""
    processor = info.context.processors.deployment
    result = await processor.create_access_token.wait_for_complete(
        action=CreateAccessTokenAction(input.to_creator())
    )
    return CreateAccessTokenPayload(access_token=AccessToken.from_dataclass(result.data))



from __future__ import annotations

import logging
import uuid
from typing import TYPE_CHECKING

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, foreign, mapped_column, relationship

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.object_storage.types import ObjectStorageData
from ai.backend.manager.models.base import (
    GUID,
    Base,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.association_artifacts_storages import (
        AssociationArtifactsStorageRow,
    )
    from ai.backend.manager.models.storage_namespace import StorageNamespaceRow

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = ("ObjectStorageRow",)


def _get_object_storage_association_artifact_join_cond() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.association_artifacts_storages import (
        AssociationArtifactsStorageRow,
    )

    return ObjectStorageRow.id == foreign(AssociationArtifactsStorageRow.storage_namespace_id)


def _get_object_storage_namespace_join_cond() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.storage_namespace import StorageNamespaceRow

    return foreign(StorageNamespaceRow.storage_id) == ObjectStorageRow.id


class ObjectStorageRow(Base):  # type: ignore[misc]
    """
    Represents an object storage configuration.
    This model is used to store the details of object storage services
    such as access keys, endpoints.
    """

    __tablename__ = "object_storages"

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    name: Mapped[str] = mapped_column("name", sa.String, index=True, unique=True, nullable=False)
    host: Mapped[str] = mapped_column("host", sa.String, index=True, nullable=False)
    access_key: Mapped[str] = mapped_column(
        "access_key",
        sa.String,
        nullable=False,
    )
    secret_key: Mapped[str] = mapped_column(
        "secret_key",
        sa.String,
        nullable=False,
    )
    endpoint: Mapped[str] = mapped_column(
        "endpoint",
        sa.String,
        nullable=False,
    )
    region: Mapped[str | None] = mapped_column(
        "region",
        sa.String,
        nullable=True,
    )

    association_artifacts_storages_rows: Mapped[list[AssociationArtifactsStorageRow]] = (
        relationship(
            "AssociationArtifactsStorageRow",
            back_populates="object_storage_row",
            primaryjoin=_get_object_storage_association_artifact_join_cond,
            overlaps="vfs_storage_row",
        )
    )
    namespace_rows: Mapped[list[StorageNamespaceRow]] = relationship(
        "StorageNamespaceRow",
        back_populates="object_storage_row",
        primaryjoin=_get_object_storage_namespace_join_cond,
    )

    def __str__(self) -> str:
        return (
            f"ObjectStorageRow("
            f"id={self.id}, "
            f"name={self.name}, "
            f"host={self.host}, "
            f"access_key={self.access_key}, "
            f"secret_key={self.secret_key}, "
            f"endpoint={self.endpoint}, "
            f"region={self.region})"
        )

    def __repr__(self) -> str:
        return self.__str__()

    def to_dataclass(self) -> ObjectStorageData:
        return ObjectStorageData(
            id=self.id,
            name=self.name,
            host=self.host,
            access_key=self.access_key,
            secret_key=self.secret_key,
            endpoint=self.endpoint,
            region=self.region,
        )



"""CSV exporter for data export."""

from __future__ import annotations

import csv
import io
import json
from collections.abc import AsyncIterator
from typing import Any

from ai.backend.manager.repositories.base.export import ExportDataStream


class CSVExporter:
    """Converts ExportDataStream to CSV bytes stream."""

    _data_stream: ExportDataStream
    _encoding: str

    def __init__(
        self,
        data_stream: ExportDataStream,
        encoding: str = "utf-8",
    ) -> None:
        self._data_stream = data_stream
        self._encoding = encoding

    async def export(self) -> AsyncIterator[bytes]:
        """Export data as CSV bytes stream.

        Yields:
            bytes: CSV data chunks including BOM (for UTF-8), header, and data rows
        """
        # UTF-8 BOM for Excel compatibility (only for UTF-8)
        if self._encoding.lower().replace("-", "") == "utf8":
            yield b"\xef\xbb\xbf"

        # CSV header
        output = io.StringIO()
        writer = csv.writer(output)
        writer.writerow(self._data_stream.field_names)
        yield output.getvalue().encode(self._encoding)

        # CSV rows (streaming per partition)
        async for partition in self._data_stream.reader:
            output = io.StringIO()
            writer = csv.writer(output)
            for row_values in partition:
                row = [self._format_value(v) for v in row_values]
                writer.writerow(row)
            yield output.getvalue().encode(self._encoding)

    def _format_value(self, value: Any) -> str:
        """Convert value to CSV string."""
        if value is None:
            return ""
        if isinstance(value, (dict, list)):
            return json.dumps(value, default=str, ensure_ascii=False)
        return str(value)



"""DomainV2 GraphQL nested types for structured field groups."""

from __future__ import annotations

from datetime import datetime

import strawberry

# ============================================================================
# Basic Information
# ============================================================================


@strawberry.type(
    name="DomainBasicInfo",
    description=(
        "Added in 26.2.0. Basic domain information. "
        "Contains identity and descriptive fields for the domain."
    ),
)
class DomainBasicInfoGQL:
    """Basic domain information."""

    name: str = strawberry.field(description="Domain name (primary key).")
    description: str | None = strawberry.field(description="Optional description of the domain.")
    integration_id: str | None = strawberry.field(
        description="External system integration identifier."
    )


# ============================================================================
# Container Registry Configuration
# ============================================================================


@strawberry.type(
    name="DomainRegistryInfo",
    description=(
        "Added in 26.2.0. Domain container registry configuration. "
        "Contains allowed container registry URLs for this domain."
    ),
)
class DomainRegistryInfoGQL:
    """Domain container registry configuration."""

    allowed_docker_registries: list[str] = strawberry.field(
        description=(
            "List of allowed container registry URLs. "
            "Empty list means no restrictions on registry access."
        )
    )


# ============================================================================
# Lifecycle Information
# ============================================================================


@strawberry.type(
    name="DomainLifecycleInfo",
    description=(
        "Added in 26.2.0. Domain lifecycle information. "
        "Contains activation status and timestamp tracking."
    ),
)
class DomainLifecycleInfoGQL:
    """Domain lifecycle information."""

    is_active: bool = strawberry.field(
        description=(
            "Whether the domain is active. "
            "Inactive domains cannot create new projects or perform operations."
        )
    )
    created_at: datetime = strawberry.field(description="Timestamp when the domain was created.")
    modified_at: datetime = strawberry.field(
        description="Timestamp when the domain was last modified."
    )



"""
WebSocket-based streaming kernel interaction APIs.

NOTE: For nginx-based setups, we need to gather all websocket-based API handlers
      under this "/stream/"-prefixed app.
"""

from __future__ import annotations

import asyncio
import base64
import json
import logging
import secrets
import uuid
import weakref
from collections import defaultdict
from collections.abc import AsyncIterator, Callable, Iterable, Mapping, MutableMapping
from datetime import timedelta
from typing import (
    TYPE_CHECKING,
    Any,
)
from urllib.parse import urlparse

import aiohttp
import aiohttp_cors
import aiotools
import attrs
import trafaret as t
import zmq
import zmq.asyncio
from aiohttp import web
from aiotools import adefer, apartial
from etcd_client import GRPCStatusCode, GRPCStatusError

from ai.backend.common import validators as tx
from ai.backend.common.events.event_types.kernel.broadcast import KernelTerminatingBroadcastEvent
from ai.backend.common.exception import BackendAIError
from ai.backend.common.json import dump_json, load_json
from ai.backend.common.types import AccessKey, AgentId, KernelId, SessionId
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.defs import DEFAULT_ROLE
from ai.backend.manager.errors.api import InvalidAPIParameters
from ai.backend.manager.errors.kernel import (
    InvalidStreamMode,
    SessionNotFound,
    TooManySessionsMatched,
)
from ai.backend.manager.errors.resource import AppNotFound, NoCurrentTaskContext
from ai.backend.manager.errors.service import AppServiceStartFailed
from ai.backend.manager.idle import AppStreamingStatus
from ai.backend.manager.models.kernel import KernelRow
from ai.backend.manager.models.session import KernelLoadingStrategy, SessionRow

from .auth import auth_required
from .manager import READ_ALLOWED, server_status_required
from .types import CORSOptions, WebMiddleware
from .utils import call_non_bursty, check_api_params
from .wsproxy import TCPProxy

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@server_status_required(READ_ALLOWED)
@auth_required
@adefer
async def stream_pty(
    defer: Callable[[Callable[[], None]], None], request: web.Request
) -> web.StreamResponse:
    root_ctx: RootContext = request.app["_root.context"]
    app_ctx: PrivateContext = request.app["stream.context"]
    database_ptask_group: aiotools.PersistentTaskGroup = request.app["database_ptask_group"]
    session_name = request.match_info["session_name"]
    access_key = request["keypair"]["access_key"]
    api_version = request["api_version"]
    try:
        async with root_ctx.db.begin_readonly_session() as db_sess:
            session = await asyncio.shield(
                database_ptask_group.create_task(
                    SessionRow.get_session(
                        db_sess,
                        session_name,
                        access_key,
                        kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
                    )
                ),
            )
    except SessionNotFound:
        raise
    log.info("STREAM_PTY(ak:{0}, s:{1})", access_key, session_name)
    compute_session: KernelRow = session.main_kernel
    stream_key = compute_session.id

    await asyncio.shield(
        database_ptask_group.create_task(
            root_ctx.registry.increment_session_usage(session),
        )
    )
    ws = web.WebSocketResponse(max_msg_size=root_ctx.config_provider.config.manager.max_wsmsg_size)
    await ws.prepare(request)

    myself = asyncio.current_task()
    if myself is None:
        raise NoCurrentTaskContext("No current task context")
    app_ctx.stream_pty_handlers[stream_key].add(myself)
    defer(lambda: app_ctx.stream_pty_handlers[stream_key].discard(myself))

    async def connect_streams(
        compute_session: KernelRow,
    ) -> tuple[zmq.asyncio.Socket, zmq.asyncio.Socket]:
        # TODO: refactor as custom row/table method
        if compute_session.kernel_host is None:
            hostname = urlparse(compute_session.agent_addr).hostname
            kernel_host = hostname.decode() if isinstance(hostname, bytes) else hostname
        else:
            kernel_host = compute_session.kernel_host
        stdin_addr = f"tcp://{kernel_host}:{compute_session.repl_in_port}"
        log.debug("stream_pty({0}): stdin: {1}", stream_key, stdin_addr)
        stdin_sock = app_ctx.zctx.socket(zmq.PUB)
        stdin_sock.connect(stdin_addr)
        stdin_sock.setsockopt(zmq.LINGER, 100)
        stdout_addr = f"tcp://{kernel_host}:{compute_session.repl_out_port}"
        log.debug("stream_pty({0}): stdout: {1}", stream_key, stdout_addr)
        stdout_sock = app_ctx.zctx.socket(zmq.SUB)
        stdout_sock.connect(stdout_addr)
        stdout_sock.setsockopt(zmq.LINGER, 100)
        stdout_sock.subscribe(b"")
        return stdin_sock, stdout_sock

    # Wrap sockets in a list so that below coroutines can share reference changes.
    socks = list(await connect_streams(compute_session))
    app_ctx.stream_stdin_socks[stream_key].add(socks[0])
    defer(lambda: app_ctx.stream_stdin_socks[stream_key].discard(socks[0]))
    stream_sync = asyncio.Event()

    async def stream_stdin() -> None:
        nonlocal socks
        try:
            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    data = load_json(msg.data)
                    if data["type"] == "stdin":
                        raw_data = base64.b64decode(data["chars"].encode("ascii"))
                        try:
                            await socks[0].send_multipart([raw_data])
                        except (RuntimeError, zmq.error.ZMQError):
                            # when socks[0] is closed, re-initiate the connection.
                            app_ctx.stream_stdin_socks[stream_key].discard(socks[0])
                            socks[1].close()
                            stdin_sock, stdout_sock = await connect_streams(compute_session)
                            socks[0] = stdin_sock
                            socks[1] = stdout_sock
                            app_ctx.stream_stdin_socks[stream_key].add(socks[0])
                            await socks[0].send_multipart([raw_data])
                            log.debug("stream_stdin({0}): zmq stream reset", stream_key)
                            stream_sync.set()
                            continue
                    else:
                        await asyncio.shield(
                            database_ptask_group.create_task(
                                root_ctx.registry.increment_session_usage(session),
                            ),
                        )
                        run_id = secrets.token_hex(8)
                        if data["type"] == "resize":
                            code = f"%resize {data['rows']} {data['cols']}"
                            await root_ctx.registry.execute(
                                session,
                                api_version,
                                run_id,
                                "query",
                                code,
                                {},
                                flush_timeout=None,
                            )
                        elif data["type"] == "ping":
                            await root_ctx.registry.execute(
                                session,
                                api_version,
                                run_id,
                                "query",
                                "%ping",
                                {},
                                flush_timeout=None,
                            )
                        elif data["type"] == "restart":
                            # Close existing zmq sockets and let stream
                            # handlers get a new one with changed stdin/stdout
                            # ports.
                            log.debug("stream_stdin: restart requested")
                            if not socks[0].closed:
                                await asyncio.shield(
                                    database_ptask_group.create_task(
                                        root_ctx.registry.restart_session(
                                            session,
                                        ),
                                    ),
                                )
                                socks[0].close()
                            else:
                                log.warning(
                                    "stream_stdin({0}): "
                                    "duplicate kernel restart request; "
                                    "ignoring it.",
                                    stream_key,
                                )
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    log.warning(
                        "stream_stdin({0}): connection closed ({1})", stream_key, ws.exception()
                    )
        except asyncio.CancelledError:
            # Agent or kernel is terminated.
            raise
        except Exception:
            await root_ctx.error_monitor.capture_exception(
                context={"user": request["user"]["uuid"]}
            )
            log.exception("stream_stdin({0}): unexpected error", stream_key)
        finally:
            log.debug("stream_stdin({0}): terminated", stream_key)
            if not socks[0].closed:
                socks[0].close()

    async def stream_stdout() -> None:
        nonlocal socks
        log.debug("stream_stdout({0}): started", stream_key)
        data: list[bytes] = []
        try:
            while True:
                try:
                    data = await socks[1].recv_multipart()
                except (asyncio.CancelledError, zmq.error.ZMQError):
                    if socks[0] not in app_ctx.stream_stdin_socks:
                        # we are terminating
                        return
                    # connection is closed, so wait until stream_stdin() recovers it.
                    await stream_sync.wait()
                    stream_sync.clear()
                    log.debug("stream_stdout({0}): zmq stream reset", stream_key)
                    continue
                if ws.closed:
                    break
                await ws.send_bytes(
                    dump_json({
                        "type": "out",
                        "data": base64.b64encode(data[0]).decode("ascii"),
                    })
                )
        except asyncio.CancelledError:
            pass
        except Exception:
            await root_ctx.error_monitor.capture_exception(
                context={"user": request["user"]["uuid"]}
            )
            log.exception("stream_stdout({0}): unexpected error", stream_key)
        finally:
            log.debug("stream_stdout({0}): terminated", stream_key)
            socks[1].close()

    # According to aiohttp docs, reading ws must be done inside this task.
    # We execute the stdout handler as another task.
    stdout_task = asyncio.create_task(stream_stdout())
    try:
        await stream_stdin()
    except Exception:
        await root_ctx.error_monitor.capture_exception(context={"user": request["user"]["uuid"]})
        log.exception("stream_pty({0}): unexpected error", stream_key)
    finally:
        stdout_task.cancel()
        await stdout_task
    return ws


@server_status_required(READ_ALLOWED)
@auth_required
@adefer
async def stream_execute(
    defer: Callable[[Callable[[], None]], None], request: web.Request
) -> web.StreamResponse:
    """
    WebSocket-version of gateway.kernel.execute().
    """
    root_ctx: RootContext = request.app["_root.context"]
    app_ctx: PrivateContext = request.app["stream.context"]
    database_ptask_group: aiotools.PersistentTaskGroup = request.app["database_ptask_group"]
    rpc_ptask_group: aiotools.PersistentTaskGroup = request.app["rpc_ptask_group"]

    config = root_ctx.config_provider.config
    registry = root_ctx.registry
    session_name = request.match_info["session_name"]
    access_key = request["keypair"]["access_key"]
    api_version = request["api_version"]
    log.info("STREAM_EXECUTE(ak:{0}, s:{1})", access_key, session_name)
    try:
        async with root_ctx.db.begin_readonly_session() as db_sess:
            session: SessionRow = await asyncio.shield(
                database_ptask_group.create_task(
                    SessionRow.get_session(
                        db_sess,
                        session_name,
                        access_key,
                        kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
                    ),
                ),
            )
    except SessionNotFound:
        raise
    stream_key = session.main_kernel.id

    await asyncio.shield(
        database_ptask_group.create_task(
            registry.increment_session_usage(session),
        )
    )
    ws = web.WebSocketResponse(max_msg_size=config.manager.max_wsmsg_size)
    await ws.prepare(request)

    myself = asyncio.current_task()
    if myself is None:
        raise NoCurrentTaskContext("No current task context")
    app_ctx.stream_execute_handlers[stream_key].add(myself)
    defer(lambda: app_ctx.stream_execute_handlers[stream_key].discard(myself))

    # This websocket connection itself is a "run".
    run_id = secrets.token_hex(8)

    try:
        if ws.closed:
            log.debug("STREAM_EXECUTE: client disconnected (cancelled)")
            return ws
        params = await ws.receive_json()
        if not params.get("mode"):
            raise InvalidStreamMode("mode is missing or empty!")
        mode = params["mode"]
        if mode not in {"query", "batch"}:
            raise InvalidStreamMode("mode has an invalid value.")
        code = params.get("code", "")
        opts = params.get("options", None) or {}

        while True:
            # TODO: rewrite agent and kernel-runner for unbuffered streaming.
            raw_result = await registry.execute(
                session, api_version, run_id, mode, code, opts, flush_timeout=0.2
            )
            if ws.closed:
                log.debug("STREAM_EXECUTE: client disconnected (interrupted)")  # type: ignore[unreachable]
                await asyncio.shield(
                    rpc_ptask_group.create_task(
                        registry.interrupt_session(session),
                    )
                )
                break
            await ws.send_json({
                "status": raw_result["status"],
                "console": raw_result.get("console"),
                "exitCode": raw_result.get("exitCode"),
                "options": raw_result.get("options"),
                "files": raw_result.get("files"),
            })
            if raw_result["status"] == "waiting-input":
                mode = "input"
                code = await ws.receive_str()
            elif raw_result["status"] == "finished":
                break
            else:
                # repeat until we get finished
                mode = "continue"
                code = ""
                opts.clear()
    except (json.decoder.JSONDecodeError, AssertionError) as e:
        log.warning("STREAM_EXECUTE: invalid/missing parameters: {0!r}", e)
        if not ws.closed:
            await ws.send_json({
                "status": "error",
                "msg": f"Invalid API parameters: {e!r}",
            })
    except BackendAIError as e:
        log.exception("STREAM_EXECUTE: exception")
        if not ws.closed:
            await ws.send_json({
                "status": "error",
                "msg": f"BackendError: {e!r}",
            })
        raise
    except asyncio.CancelledError:
        if not ws.closed:
            await ws.send_json({
                "status": "server-restarting",
                "msg": (
                    "The API server is going to restart for maintenance. "
                    "Please connect again with the same run ID."
                ),
            })
        raise
    finally:
        return ws


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(["app", "service"]): t.String,
        # The port argument is only required to use secondary ports
        # when the target app listens multiple TCP ports.
        # Otherwise it should be omitted or set to the same value of
        # the actual port number used by the app.
        tx.AliasedKey(["port"], default=None): t.Null | t.Int[1024:65535],
        tx.AliasedKey(["envs"], default=None): t.Null | t.String,  # stringified JSON
        # e.g., '{"PASSWORD": "12345"}'
        tx.AliasedKey(["arguments"], default=None): t.Null | t.String,  # stringified JSON
        # e.g., '{"-P": "12345"}'
        # The value can be one of:
        # None, str, List[str]
    })
)
@adefer
async def stream_proxy(
    defer: Callable[[Callable[[], None]], None], request: web.Request, params: Mapping[str, Any]
) -> web.StreamResponse:
    root_ctx: RootContext = request.app["_root.context"]
    app_ctx: PrivateContext = request.app["stream.context"]
    database_ptask_group: aiotools.PersistentTaskGroup = request.app["database_ptask_group"]
    rpc_ptask_group: aiotools.PersistentTaskGroup = request.app["rpc_ptask_group"]
    session_name: str = request.match_info["session_name"]
    access_key: AccessKey = request["keypair"]["access_key"]
    service: str = params["app"]
    myself = asyncio.current_task()
    if myself is None:
        raise NoCurrentTaskContext("No current task context")
    try:
        async with root_ctx.db.begin_readonly_session() as db_sess:
            session = await asyncio.shield(
                database_ptask_group.create_task(
                    SessionRow.get_session(
                        db_sess,
                        session_name,
                        access_key,
                        kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
                    ),
                )
            )
    except (SessionNotFound, TooManySessionsMatched):
        raise
    kernel: KernelRow = session.main_kernel
    kernel_id = kernel.id
    session_id = SessionId(session.id)
    stream_key = kernel_id
    stream_id = uuid.uuid4().hex
    app_ctx.stream_proxy_handlers[stream_key].add(myself)
    defer(lambda: app_ctx.stream_proxy_handlers[stream_key].discard(myself))
    kernel_host: str
    if kernel.kernel_host is None:
        hostname = urlparse(kernel.agent_addr).hostname
        if hostname is None:
            raise InvalidAPIParameters(
                f"Cannot determine kernel host from agent address: {kernel.agent_addr}"
            )
        kernel_host = hostname.decode() if isinstance(hostname, bytes) else hostname
    else:
        kernel_host = kernel.kernel_host
    service_ports: list[dict[str, Any]] = kernel.service_ports or []
    sport: dict[str, Any] = {}
    host_port: int = 0
    dest: tuple[str, int] = ("", 0)
    for sport in service_ports:
        if sport["name"] == service:
            if params["port"]:
                # using one of the primary/secondary ports of the app
                try:
                    hport_idx = sport["container_ports"].index(params["port"])
                except ValueError as e:
                    raise InvalidAPIParameters(
                        f"Service {service} does not open the port number {params['port']}."
                    ) from e
                host_port = sport["host_ports"][hport_idx]
            else:  # using the default (primary) port of the app
                if "host_ports" not in sport:
                    host_port = sport["host_port"]  # legacy kernels
                else:
                    host_port = sport["host_ports"][0]
            dest = (kernel_host, host_port)
            break
    else:
        raise AppNotFound(f"{session_name}:{service}")

    log.info(
        "STREAM_WSPROXY (ak:{}, s:{}): tunneling {}:{} to {}",
        access_key,
        session_name,
        service,
        sport["protocol"],
        "{}:{}".format(*dest),
    )
    if sport["protocol"] == "tcp":
        proxy_cls = TCPProxy
    elif sport["protocol"] == "pty":
        raise NotImplementedError
    elif sport["protocol"] in ("http", "preopen", "vnc", "rdp"):
        proxy_cls = TCPProxy
    else:
        raise InvalidAPIParameters(f"Unsupported service protocol: {sport['protocol']}")

    valkey_live = root_ctx.valkey_live
    conn_tracker_key = f"session.{kernel_id}.active_app_connections"

    async def update_connection_tracker() -> None:
        """Update connection tracker with current timestamp."""
        await valkey_live.update_app_connection_tracker(str(kernel_id), service, stream_id)

    async def refresh_cb(_kernel_id_str: str, _data: bytes) -> None:
        await asyncio.shield(
            rpc_ptask_group.create_task(
                call_non_bursty(
                    conn_tracker_key,
                    update_connection_tracker,
                    max_bursts=128,
                    max_idle=5000,
                ),
            )
        )

    down_cb = apartial(refresh_cb, str(kernel_id))
    up_cb = apartial(refresh_cb, str(kernel_id))
    ping_cb = apartial(refresh_cb, str(kernel_id))

    async def add_conn_track() -> None:
        async with app_ctx.conn_tracker_lock:
            app_ctx.active_session_ids[kernel_id] += 1
            await valkey_live.update_connection_tracker(str(kernel_id), service, stream_id)
            await root_ctx.idle_checker_host.update_app_streaming_status(
                session_id,
                AppStreamingStatus.HAS_ACTIVE_CONNECTIONS,
            )

    async def clear_conn_track() -> None:
        async with app_ctx.conn_tracker_lock:
            app_ctx.active_session_ids[kernel_id] -= 1
            if app_ctx.active_session_ids[kernel_id] <= 0:
                del app_ctx.active_session_ids[kernel_id]
            await valkey_live.remove_connection_tracker(str(kernel_id), service, stream_id)
            remaining_count = await valkey_live.count_active_connections(str(kernel_id))
            if remaining_count == 0:
                await root_ctx.idle_checker_host.update_app_streaming_status(
                    session_id,
                    AppStreamingStatus.NO_ACTIVE_CONNECTIONS,
                )

    try:
        await asyncio.shield(
            database_ptask_group.create_task(
                add_conn_track(),
            )
        )
        await asyncio.shield(
            database_ptask_group.create_task(
                root_ctx.registry.increment_session_usage(session),
            )
        )

        opts: MutableMapping[str, None | str | list[str]] = {}
        if params["arguments"] is not None:
            opts["arguments"] = load_json(params["arguments"])
        if params["envs"] is not None:
            opts["envs"] = load_json(params["envs"])

        result = await asyncio.shield(
            rpc_ptask_group.create_task(
                root_ctx.registry.start_service(session, service, opts),
            ),
        )
        if result["status"] == "failed":
            raise AppServiceStartFailed(
                "Failed to launch the app service", extra_data=result["error"]
            )

        # TODO: weakref to proxies for graceful shutdown?
        ws = web.WebSocketResponse(
            autoping=False,
            max_msg_size=root_ctx.config_provider.config.manager.max_wsmsg_size,
        )
        await ws.prepare(request)
        proxy = proxy_cls(
            ws,
            dest[0],
            dest[1],
            downstream_callback=down_cb,
            upstream_callback=up_cb,
            ping_callback=ping_cb,
        )
        return await proxy.proxy()
    except asyncio.CancelledError:
        log.debug("stream_proxy({}, {}) cancelled", stream_key, service)
        raise
    finally:
        await asyncio.shield(database_ptask_group.create_task(clear_conn_track()))


@server_status_required(READ_ALLOWED)
@auth_required
async def get_stream_apps(request: web.Request) -> web.Response:
    session_name = request.match_info["session_name"]
    access_key = request["keypair"]["access_key"]
    root_ctx: RootContext = request.app["_root.context"]
    async with root_ctx.db.begin_readonly_session() as db_sess:
        compute_session = await SessionRow.get_session(
            db_sess,
            session_name,
            access_key,
            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
        )
    raw_service_ports = compute_session.main_kernel.service_ports
    if raw_service_ports is None:
        return web.json_response([])
    service_ports: list[dict[str, Any]] = raw_service_ports
    resp = []
    for item in service_ports:
        response_dict: dict[str, Any] = {
            "name": item["name"],
            "protocol": item["protocol"],
            "ports": item["container_ports"],
        }
        if "url_template" in item.keys():
            response_dict["url_template"] = item["url_template"]
        if "allowed_arguments" in item.keys():
            response_dict["allowed_arguments"] = item["allowed_arguments"]
        if "allowed_envs" in item.keys():
            response_dict["allowed_envs"] = item["allowed_envs"]
        resp.append(response_dict)
    return web.json_response(resp)


async def handle_kernel_terminating(
    app: web.Application,
    _source: AgentId,
    event: KernelTerminatingBroadcastEvent,
) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["stream.context"]
    try:
        kernel = await KernelRow.get_kernel(
            root_ctx.db,
            event.kernel_id,
            allow_stale=True,
        )
    except SessionNotFound:
        return
    if kernel.cluster_role == DEFAULT_ROLE:
        stream_key = kernel.id
        cancelled_tasks = []
        for sock in app_ctx.stream_stdin_socks[stream_key]:
            sock.close()
        for handler in list(app_ctx.stream_pty_handlers.get(stream_key, [])):
            handler.cancel()
            cancelled_tasks.append(handler)
        for handler in list(app_ctx.stream_execute_handlers.get(stream_key, [])):
            handler.cancel()
            cancelled_tasks.append(handler)
        for handler in list(app_ctx.stream_proxy_handlers.get(stream_key, [])):
            handler.cancel()
            cancelled_tasks.append(handler)
        await asyncio.gather(*cancelled_tasks, return_exceptions=True)
        # TODO: reconnect if restarting?


async def stream_conn_tracker_gc(root_ctx: RootContext, app_ctx: PrivateContext) -> None:
    valkey_live = root_ctx.valkey_live
    try:
        while True:
            try:
                # TODO: Use `no_packet_timeout` from the shared config after resolving type issue.
                no_packet_timeout: timedelta = tx.TimeDuration().check(
                    await root_ctx.etcd.get("config/idle/app-streaming-packet-timeout") or "5m",
                )
            except GRPCStatusError as e:
                err_detail = e.args[0]

                if err_detail["code"] == GRPCStatusCode.Unavailable:
                    log.warning(
                        "stream_conn_tracker_gc(): error while connecting to Etcd server,"
                        " retrying..."
                    )
                    continue
                raise e
            async with app_ctx.conn_tracker_lock:
                now = await valkey_live.get_server_time()
                for session_id in app_ctx.active_session_ids.keys():
                    prev_remaining_count = await valkey_live.count_active_connections(
                        str(session_id)
                    )
                    removed_count = await valkey_live.remove_stale_connections(
                        str(session_id),
                        now - no_packet_timeout.total_seconds(),
                    )
                    remaining_count = await valkey_live.count_active_connections(str(session_id))
                    log.debug(
                        f"conn_tracker: gc {session_id} "
                        f"removed/remaining = {removed_count}/{remaining_count}",
                    )
                    if prev_remaining_count > 0 and remaining_count == 0:
                        # Note: kernel_id is used as session_id key here for connection tracking
                        # The idle checker operates on session granularity
                        await root_ctx.idle_checker_host.update_app_streaming_status(
                            SessionId(session_id),
                            AppStreamingStatus.NO_ACTIVE_CONNECTIONS,
                        )
            await asyncio.sleep(10)
    except asyncio.CancelledError:
        pass


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    stream_pty_handlers: defaultdict[KernelId, weakref.WeakSet[asyncio.Task[Any]]]
    stream_execute_handlers: defaultdict[KernelId, weakref.WeakSet[asyncio.Task[Any]]]
    stream_proxy_handlers: defaultdict[KernelId, weakref.WeakSet[asyncio.Task[Any]]]
    stream_stdin_socks: defaultdict[KernelId, weakref.WeakSet[zmq.asyncio.Socket]]
    zctx: zmq.asyncio.Context
    conn_tracker_lock: asyncio.Lock
    conn_tracker_gc_task: asyncio.Task[Any]
    active_session_ids: defaultdict[KernelId, int]


async def stream_app_ctx(app: web.Application) -> AsyncIterator[None]:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["stream.context"]

    app_ctx.stream_pty_handlers = defaultdict(weakref.WeakSet)
    app_ctx.stream_execute_handlers = defaultdict(weakref.WeakSet)
    app_ctx.stream_proxy_handlers = defaultdict(weakref.WeakSet)
    app_ctx.stream_stdin_socks = defaultdict(weakref.WeakSet)
    app_ctx.zctx = zmq.asyncio.Context()
    app_ctx.conn_tracker_lock = asyncio.Lock()
    app_ctx.active_session_ids = defaultdict(int)  # multiset[int]
    app_ctx.conn_tracker_gc_task = asyncio.create_task(stream_conn_tracker_gc(root_ctx, app_ctx))

    root_ctx.event_dispatcher.subscribe(
        KernelTerminatingBroadcastEvent, app, handle_kernel_terminating
    )

    yield

    # The shutdown handler below is called before this cleanup.
    app_ctx.zctx.term()


async def stream_shutdown(app: web.Application) -> None:
    database_ptask_group: aiotools.PersistentTaskGroup = app["database_ptask_group"]
    rpc_ptask_group: aiotools.PersistentTaskGroup = app["rpc_ptask_group"]
    await database_ptask_group.shutdown()
    await rpc_ptask_group.shutdown()
    cancelled_tasks: list[asyncio.Task[Any]] = []
    app_ctx: PrivateContext = app["stream.context"]
    app_ctx.conn_tracker_gc_task.cancel()
    cancelled_tasks.append(app_ctx.conn_tracker_gc_task)
    for per_kernel_handlers in app_ctx.stream_pty_handlers.values():
        for handler in list(per_kernel_handlers):
            if not handler.done():
                handler.cancel()
                cancelled_tasks.append(handler)
    for per_kernel_handlers in app_ctx.stream_execute_handlers.values():
        for handler in list(per_kernel_handlers):
            if not handler.done():
                handler.cancel()
                cancelled_tasks.append(handler)
    for per_kernel_handlers in app_ctx.stream_proxy_handlers.values():
        for handler in list(per_kernel_handlers):
            if not handler.done():
                handler.cancel()
                cancelled_tasks.append(handler)
    await asyncio.gather(*cancelled_tasks, return_exceptions=True)


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.cleanup_ctx.append(stream_app_ctx)
    app.on_shutdown.append(stream_shutdown)
    app["prefix"] = "stream"
    app["api_versions"] = (2, 3, 4)
    app["stream.context"] = PrivateContext()
    app["database_ptask_group"] = aiotools.PersistentTaskGroup()
    app["rpc_ptask_group"] = aiotools.PersistentTaskGroup()
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    cors.add(add_route("GET", r"/session/{session_name}/pty", stream_pty))
    cors.add(add_route("GET", r"/session/{session_name}/execute", stream_execute))
    cors.add(add_route("GET", r"/session/{session_name}/apps", get_stream_apps))
    # internally both tcp/http proxies use websockets as API/agent-level transports,
    # and thus they have the same implementation here.
    cors.add(add_route("GET", r"/session/{session_name}/httpproxy", stream_proxy))
    cors.add(add_route("GET", r"/session/{session_name}/tcpproxy", stream_proxy))
    return app, []



"""Resource Slot Repositories wrapper."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.types import RepositoryArgs

from .repository import ResourceSlotRepository

__all__ = ("ResourceSlotRepositories",)


@dataclass
class ResourceSlotRepositories:
    repository: ResourceSlotRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = ResourceSlotRepository(args.db)

        return cls(
            repository=repository,
        )



from __future__ import annotations

import uuid
from dataclasses import dataclass

from ai.backend.common.dto.manager.response import ObjectStorageResponse


@dataclass
class ObjectStorageListResult:
    """Search result with total count for Object storages."""

    items: list[ObjectStorageData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


@dataclass
class ObjectStorageData:
    id: uuid.UUID
    name: str
    host: str
    access_key: str
    secret_key: str
    endpoint: str
    region: str | None

    def to_dto(self) -> ObjectStorageResponse:
        return ObjectStorageResponse(
            id=str(self.id),
            name=self.name,
            host=self.host,
            access_key=self.access_key,
            secret_key=self.secret_key,
            endpoint=self.endpoint,
            region=self.region or "",
        )



from .drf import DRFSequencer
from .fair_share import FairShareSequencer
from .fifo import FIFOSequencer
from .lifo import LIFOSequencer
from .sequencer import WorkloadSequencer

__all__ = [
    "DRFSequencer",
    "FairShareSequencer",
    "FIFOSequencer",
    "LIFOSequencer",
    "WorkloadSequencer",
]



from __future__ import annotations

import logging
import uuid
from collections.abc import Container, Iterable, Sequence
from dataclasses import dataclass, field
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Self,
    TypedDict,
    overload,
    override,
)

import sqlalchemy as sa
import trafaret as t
from sqlalchemy.exc import NoResultFound
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import (
    Mapped,
    foreign,
    joinedload,
    load_only,
    mapped_column,
    relationship,
    selectinload,
)
from sqlalchemy.orm.strategy_options import _AbstractLoad

from ai.backend.common import msgpack
from ai.backend.common.types import ResourceSlot, VFolderHostPermissionMap
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.group.types import GroupData, ProjectType
from ai.backend.manager.defs import RESERVED_DOTFILES
from ai.backend.manager.errors.common import ObjectNotFound
from ai.backend.manager.models.association_container_registries_groups import (
    AssociationContainerRegistriesGroupsRow,
)
from ai.backend.manager.models.base import (
    GUID,
    Base,
    EnumValueType,
    ResourceSlotColumn,
    SlugType,
    StructuredJSONColumn,
    VFolderHostPermissionColumn,
)
from ai.backend.manager.models.rbac import (
    AbstractPermissionContext,
    AbstractPermissionContextBuilder,
    DomainScope,
    ProjectScope,
    RBACModel,
    ScopeType,
    UserScope,
    get_predefined_roles_in_scope,
    required_permission,
)
from ai.backend.manager.models.rbac.context import ClientContext
from ai.backend.manager.models.rbac.permission_defs import ProjectPermission
from ai.backend.manager.models.types import (
    QueryCondition,
    QueryOption,
    load_related_field,
)
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine, execute_with_txn_retry

if TYPE_CHECKING:
    from ai.backend.manager.models.domain import DomainRow
    from ai.backend.manager.models.kernel import KernelRow
    from ai.backend.manager.models.network import NetworkRow
    from ai.backend.manager.models.rbac import ContainerRegistryScope
    from ai.backend.manager.models.resource_policy import ProjectResourcePolicyRow
    from ai.backend.manager.models.scaling_group import ScalingGroupForProjectRow
    from ai.backend.manager.models.session import SessionRow
    from ai.backend.manager.models.user import UserRow
    from ai.backend.manager.models.vfolder import VFolderRow

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


def _get_networks_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.network import NetworkRow

    return GroupRow.id == foreign(NetworkRow.project)


def _get_vfolder_rows_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.vfolder import VFolderRow

    return GroupRow.id == foreign(VFolderRow.group)


def _get_association_container_registries_groups_join_condition() -> sa.ColumnElement[bool]:
    return GroupRow.id == foreign(AssociationContainerRegistriesGroupsRow.group_id)


__all__: Sequence[str] = (
    "MAXIMUM_DOTFILE_SIZE",
    "AssocGroupUserRow",
    "GroupDotfile",
    "GroupRow",
    "ProjectType",
    "association_groups_users",
    "groups",
    "query_group_domain",
    "query_group_dotfiles",
    "resolve_group_name_or_id",
    "verify_dotfile_name",
)

MAXIMUM_DOTFILE_SIZE = 64 * 1024  # 61 KiB


container_registry_iv = t.Dict({}) | t.Dict({
    t.Key("registry"): t.String(),
    t.Key("project"): t.String(),
})


class AssocGroupUserRow(Base):  # type: ignore[misc]
    __tablename__ = "association_groups_users"
    __table_args__ = (
        sa.UniqueConstraint("user_id", "group_id", name="uq_association_user_id_group_id"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    user_id: Mapped[uuid.UUID] = mapped_column(
        "user_id",
        GUID,
        sa.ForeignKey("users.uuid", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    )
    group_id: Mapped[uuid.UUID] = mapped_column(
        "group_id",
        GUID,
        sa.ForeignKey("groups.id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    )

    user: Mapped[UserRow] = relationship("UserRow", back_populates="groups")
    group: Mapped[GroupRow] = relationship("GroupRow", back_populates="users")


# NOTE: Deprecated legacy table reference for backward compatibility.
# Use AssocGroupUserRow class directly for new code.
association_groups_users = AssocGroupUserRow.__table__


class GroupRow(Base):  # type: ignore[misc]
    __tablename__ = "groups"
    __table_args__ = (
        sa.UniqueConstraint("name", "domain_name", name="uq_groups_name_domain_name"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    name: Mapped[str] = mapped_column(
        "name", SlugType(length=64, allow_unicode=True, allow_dot=True), nullable=False
    )
    description: Mapped[str | None] = mapped_column("description", sa.String(length=512))
    is_active: Mapped[bool | None] = mapped_column("is_active", sa.Boolean, default=True)
    created_at: Mapped[datetime | None] = mapped_column(
        "created_at", sa.DateTime(timezone=True), server_default=sa.func.now()
    )
    modified_at: Mapped[datetime | None] = mapped_column(
        "modified_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        onupdate=sa.func.current_timestamp(),
    )
    #: Field for synchronization with external services.
    integration_id: Mapped[str | None] = mapped_column("integration_id", sa.String(length=512))
    domain_name: Mapped[str] = mapped_column(
        "domain_name",
        sa.String(length=64),
        sa.ForeignKey("domains.name", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    # TODO: separate resource-related fields with new domain resource policy table when needed.
    total_resource_slots: Mapped[ResourceSlot] = mapped_column(
        "total_resource_slots", ResourceSlotColumn(), default=ResourceSlot, nullable=False
    )
    allowed_vfolder_hosts: Mapped[VFolderHostPermissionMap] = mapped_column(
        "allowed_vfolder_hosts",
        VFolderHostPermissionColumn(),
        nullable=False,
        default=VFolderHostPermissionMap,
    )
    # dotfiles column, \x90 means empty list in msgpack
    dotfiles: Mapped[bytes] = mapped_column(
        "dotfiles", sa.LargeBinary(length=MAXIMUM_DOTFILE_SIZE), nullable=False, default=b"\x90"
    )
    resource_policy: Mapped[str] = mapped_column(
        "resource_policy",
        sa.String(length=256),
        sa.ForeignKey("project_resource_policies.name"),
        nullable=False,
    )
    type: Mapped[ProjectType] = mapped_column(
        "type",
        EnumValueType(ProjectType),
        nullable=False,
        default=ProjectType.GENERAL,
    )
    container_registry: Mapped[dict[str, Any] | None] = mapped_column(
        "container_registry",
        StructuredJSONColumn(container_registry_iv),
        nullable=True,
        default=None,
    )

    # Relationships (defined with deferred join conditions to avoid circular imports)
    sessions: Mapped[list[SessionRow]] = relationship("SessionRow", back_populates="group")
    domain: Mapped[DomainRow] = relationship("DomainRow", back_populates="groups")
    sgroup_for_groups_rows: Mapped[list[ScalingGroupForProjectRow]] = relationship(
        "ScalingGroupForProjectRow", back_populates="project_row"
    )
    users: Mapped[list[AssocGroupUserRow]] = relationship(
        "AssocGroupUserRow", back_populates="group"
    )
    resource_policy_row: Mapped[ProjectResourcePolicyRow] = relationship(
        "ProjectResourcePolicyRow", back_populates="projects"
    )
    kernels: Mapped[list[KernelRow]] = relationship("KernelRow", back_populates="group_row")
    networks: Mapped[list[NetworkRow]] = relationship(
        "NetworkRow",
        back_populates="project_row",
        primaryjoin=_get_networks_join_condition,
    )
    vfolder_rows: Mapped[list[VFolderRow]] = relationship(
        "VFolderRow",
        back_populates="group_row",
        primaryjoin=_get_vfolder_rows_join_condition,
    )
    association_container_registries_groups_rows: Mapped[
        list[AssociationContainerRegistriesGroupsRow]
    ] = relationship(
        "AssociationContainerRegistriesGroupsRow",
        back_populates="group_row",
        primaryjoin=_get_association_container_registries_groups_join_condition,
    )

    def to_data(self) -> GroupData:
        return GroupData(
            id=self.id,
            name=self.name,
            description=self.description,
            is_active=self.is_active,
            created_at=self.created_at,
            modified_at=self.modified_at,
            integration_id=self.integration_id,
            domain_name=self.domain_name,
            total_resource_slots=self.total_resource_slots,
            allowed_vfolder_hosts=self.allowed_vfolder_hosts,
            dotfiles=self.dotfiles,
            resource_policy=self.resource_policy,
            type=self.type,
            container_registry=self.container_registry,
        )

    @classmethod
    async def get(
        cls,
        session: AsyncSession,
        project_id: uuid.UUID,
        load_resource_policy: bool = False,
    ) -> GroupRow:
        query = sa.select(GroupRow).filter(GroupRow.id == project_id)
        if load_resource_policy:
            query = query.options(selectinload(GroupRow.resource_policy_row))
        row = await session.scalar(query)
        if not row:
            raise NoResultFound

        return row

    @classmethod
    def load_resource_policy(cls) -> _AbstractLoad:
        return joinedload(GroupRow.resource_policy_row)

    @classmethod
    async def query_by_condition(
        cls,
        conditions: Sequence[QueryCondition],
        options: Sequence[QueryOption] = tuple(),
        *,
        db: ExtendedAsyncSAEngine,
    ) -> Sequence[GroupRow]:
        """
        Args:
            condition: QueryCondition.
            options: A sequence of query options.
            db: Database engine.
        Returns:
            A list of GroupRow instances that match the condition.
        Raises:
            EmptySQLCondition: If the condition is empty.
        """
        query_stmt = sa.select(GroupRow)
        for cond in conditions:
            query_stmt = cond(query_stmt)

        for option in options:
            query_stmt = option(query_stmt)

        async def fetch(db_session: AsyncSession) -> Sequence[GroupRow]:
            return (await db_session.scalars(query_stmt)).all()

        async with db.connect() as db_conn:
            return await execute_with_txn_retry(
                fetch,
                db.begin_readonly_session,
                db_conn,
            )

    @classmethod
    async def get_by_id_with_policies(
        cls,
        project_id: uuid.UUID,
        *,
        db: ExtendedAsyncSAEngine,
    ) -> Self:
        """
        Query a project by its ID with related resource policies.
        Args:
            project_id: The ID of the project.
            db: Database engine.
        Returns:
            The GroupRow instance that matches the project ID.
        Raises:
            ObjectNotFound: If the project not found.
        """
        rows = await cls.query_by_condition(
            [by_id(project_id)],
            [load_related_field(cls.load_resource_policy())],
            db=db,
        )
        if not rows:
            raise ObjectNotFound(f"Project with id {project_id} not found")
        return rows[0]


# NOTE: Deprecated legacy table reference for backward compatibility.
# Use GroupRow class directly for new code.
groups = GroupRow.__table__


def by_id(project_id: uuid.UUID) -> QueryCondition:
    def _by_id(
        query_stmt: sa.sql.Select[Any],
    ) -> sa.sql.Select[Any]:
        return query_stmt.where(GroupRow.id == project_id)

    return _by_id


@dataclass
class ProjectModel(RBACModel[ProjectPermission]):
    id: uuid.UUID
    name: str
    description: str | None
    is_active: bool | None
    created_at: datetime | None
    modified_at: datetime | None
    domain_name: str
    type: str

    _integration_id: str | None
    _total_resource_slots: ResourceSlot
    _allowed_vfolder_hosts: VFolderHostPermissionMap
    _dotfiles: bytes
    _resource_policy: str
    _container_registry: dict[str, str] | None

    _permissions: frozenset[ProjectPermission] = field(default_factory=frozenset)

    @property
    def permissions(self) -> Container[ProjectPermission]:
        return self._permissions

    @property
    @required_permission(ProjectPermission.READ_SENSITIVE_ATTRIBUTE)
    def integration_id(self) -> str | None:
        return self._integration_id

    @property
    @required_permission(ProjectPermission.READ_SENSITIVE_ATTRIBUTE)
    def total_resource_slots(self) -> ResourceSlot:
        return self._total_resource_slots

    @property
    @required_permission(ProjectPermission.READ_SENSITIVE_ATTRIBUTE)
    def allowed_vfolder_hosts(self) -> VFolderHostPermissionMap:
        return self._allowed_vfolder_hosts

    @property
    @required_permission(ProjectPermission.READ_SENSITIVE_ATTRIBUTE)
    def dotfiles(self) -> bytes:
        return self._dotfiles

    @property
    @required_permission(ProjectPermission.READ_SENSITIVE_ATTRIBUTE)
    def resource_policy(self) -> str:
        return self._resource_policy

    @property
    @required_permission(ProjectPermission.READ_SENSITIVE_ATTRIBUTE)
    def container_registry(self) -> dict[str, Any] | None:
        return self._container_registry

    @classmethod
    def from_row(cls, row: GroupRow, permissions: Iterable[ProjectPermission]) -> Self:
        return cls(
            id=row.id,
            name=row.name,
            description=row.description,
            is_active=row.is_active,
            created_at=row.created_at,
            modified_at=row.modified_at,
            domain_name=row.domain_name,
            type=row.type,
            _integration_id=row.integration_id,
            _total_resource_slots=row.total_resource_slots,
            _allowed_vfolder_hosts=row.allowed_vfolder_hosts,
            _dotfiles=row.dotfiles,
            _resource_policy=row.resource_policy,
            _container_registry=row.container_registry,
            _permissions=frozenset(permissions),
        )


def _build_group_query(
    cond: sa.sql.expression.BinaryExpression[Any], domain_name: str
) -> sa.sql.Select[Any]:
    return (
        sa.select(groups.c.id)
        .select_from(groups)
        .where(
            cond & (groups.c.domain_name == domain_name),
        )
    )


async def resolve_group_name_or_id(
    db_conn: SAConnection,
    domain_name: str,
    value: str | uuid.UUID,
) -> uuid.UUID | None:
    match value:
        case uuid.UUID():
            cond = groups.c.id == value
        case str():
            cond = groups.c.name == value
        case _:
            raise TypeError("unexpected type for group_name_or_id")
    query = _build_group_query(cond, domain_name)
    result: uuid.UUID | None = await db_conn.scalar(query)
    return result


@overload
async def resolve_groups(
    db_conn: SAConnection,
    domain_name: str,
    values: Iterable[uuid.UUID],
) -> Iterable[uuid.UUID]: ...


@overload
async def resolve_groups(
    db_conn: SAConnection,
    domain_name: str,
    values: Iterable[str],
) -> Iterable[uuid.UUID]: ...


async def resolve_groups(
    db_conn: SAConnection,
    domain_name: str,
    values: Iterable[uuid.UUID] | Iterable[str],
) -> Iterable[uuid.UUID]:
    listed_val = [*values]
    match listed_val:
        case [uuid.UUID(), *_]:
            query = _build_group_query((groups.c.id.in_(listed_val)), domain_name)
        case [str(), *_]:
            query = _build_group_query((groups.c.name.in_(listed_val)), domain_name)
        case []:
            return []
        case _:
            raise TypeError("unexpected type for group_name_or_id")

    rows = (await db_conn.execute(query)).fetchall()
    return [row.id for row in rows]


class GroupDotfile(TypedDict):
    data: str
    path: str
    perm: str


async def query_group_dotfiles(
    db_conn: SAConnection,
    group_id: GUID[uuid.UUID] | uuid.UUID,
) -> tuple[list[GroupDotfile], int]:
    query = sa.select(groups.c.dotfiles).select_from(groups).where(groups.c.id == group_id)
    packed_dotfile = await db_conn.scalar(query)
    if packed_dotfile is None:
        return [], MAXIMUM_DOTFILE_SIZE
    rows = msgpack.unpackb(packed_dotfile)
    return rows, MAXIMUM_DOTFILE_SIZE - len(packed_dotfile)


async def query_group_domain(
    db_conn: SAConnection,
    group_id: GUID[uuid.UUID] | uuid.UUID,
) -> str | None:
    query = sa.select(groups.c.domain_name).select_from(groups).where(groups.c.id == group_id)
    return await db_conn.scalar(query)


def verify_dotfile_name(dotfile: str) -> bool:
    return dotfile not in RESERVED_DOTFILES


ALL_PROJECT_PERMISSIONS = frozenset([perm for perm in ProjectPermission])
OWNER_PERMISSIONS: frozenset[ProjectPermission] = ALL_PROJECT_PERMISSIONS
ADMIN_PERMISSIONS: frozenset[ProjectPermission] = ALL_PROJECT_PERMISSIONS
MONITOR_PERMISSIONS: frozenset[ProjectPermission] = frozenset([
    ProjectPermission.READ_ATTRIBUTE,
    ProjectPermission.READ_SENSITIVE_ATTRIBUTE,
    ProjectPermission.UPDATE_ATTRIBUTE,
])
PRIVILEGED_MEMBER_PERMISSIONS: frozenset[ProjectPermission] = frozenset([
    ProjectPermission.READ_ATTRIBUTE
])
MEMBER_PERMISSIONS: frozenset[ProjectPermission] = frozenset([ProjectPermission.READ_ATTRIBUTE])

type WhereClauseType = (
    sa.sql.expression.BinaryExpression[Any]
    | sa.sql.expression.BooleanClauseList
    | sa.sql.elements.ColumnElement[bool]
)


@dataclass
class ProjectPermissionContext(AbstractPermissionContext[ProjectPermission, GroupRow, uuid.UUID]):
    registry_id_to_additional_permission_map: dict[uuid.UUID, frozenset[ProjectPermission]] = field(
        default_factory=dict
    )

    @property
    def query_condition(self) -> WhereClauseType | None:
        cond: WhereClauseType | None = None

        def _OR_coalesce(
            base_cond: WhereClauseType | None,
            _cond: WhereClauseType,
        ) -> WhereClauseType:
            return base_cond | _cond if base_cond is not None else _cond

        if self.registry_id_to_additional_permission_map:
            registry_id = list(self.registry_id_to_additional_permission_map)[0]

            cond = _OR_coalesce(
                cond,
                GroupRow.association_container_registries_groups_rows.any(
                    AssociationContainerRegistriesGroupsRow.registry_id == registry_id
                ),
            )
        if self.domain_name_to_permission_map:
            cond = _OR_coalesce(
                cond, GroupRow.domain_name.in_(self.domain_name_to_permission_map.keys())
            )
        if self.object_id_to_additional_permission_map:
            cond = _OR_coalesce(
                cond, GroupRow.id.in_(self.object_id_to_additional_permission_map.keys())
            )
        if self.object_id_to_overriding_permission_map:
            cond = _OR_coalesce(
                cond, GroupRow.id.in_(self.object_id_to_overriding_permission_map.keys())
            )
        return cond

    async def build_query(self) -> sa.sql.Select[Any] | None:
        cond = self.query_condition
        if cond is None:
            return None
        return sa.select(GroupRow).where(cond)

    async def calculate_final_permission(self, rbac_obj: GroupRow) -> frozenset[ProjectPermission]:
        project_row = rbac_obj
        project_id = project_row.id
        permissions: frozenset[ProjectPermission] = frozenset()

        if (
            overriding_perm := self.object_id_to_overriding_permission_map.get(project_id)
        ) is not None:
            permissions = overriding_perm
        else:
            permissions |= self.object_id_to_additional_permission_map.get(project_id, set())
            permissions |= self.domain_name_to_permission_map.get(project_row.domain_name, set())
        return permissions


class ProjectPermissionContextBuilder(
    AbstractPermissionContextBuilder[ProjectPermission, ProjectPermissionContext]
):
    db_session: AsyncSession

    def __init__(self, db_session: AsyncSession) -> None:
        self.db_session = db_session

    @override
    async def calculate_permission(
        self,
        ctx: ClientContext,
        target_scope: ScopeType,
    ) -> frozenset[ProjectPermission]:
        roles = await get_predefined_roles_in_scope(ctx, target_scope, self.db_session)
        return await self._calculate_permission_by_predefined_roles(roles)

    @override
    async def build_ctx_in_system_scope(
        self,
        ctx: ClientContext,
    ) -> ProjectPermissionContext:
        from ai.backend.manager.models.domain import DomainRow

        perm_ctx = ProjectPermissionContext()
        _domain_query_stmt = sa.select(DomainRow).options(load_only(DomainRow.name))
        for row in await self.db_session.scalars(_domain_query_stmt):
            to_be_merged = await self.build_ctx_in_domain_scope(ctx, DomainScope(row.name))
            perm_ctx.merge(to_be_merged)
        return perm_ctx

    @override
    async def build_ctx_in_domain_scope(
        self,
        ctx: ClientContext,
        scope: DomainScope,
    ) -> ProjectPermissionContext:
        permissions = await self.calculate_permission(ctx, scope)
        return ProjectPermissionContext(
            domain_name_to_permission_map={scope.domain_name: permissions}
        )

    @override
    async def build_ctx_in_project_scope(
        self, ctx: ClientContext, scope: ProjectScope
    ) -> ProjectPermissionContext:
        permissions = await self.calculate_permission(ctx, scope)
        return ProjectPermissionContext(
            object_id_to_additional_permission_map={scope.project_id: permissions}
        )

    @override
    async def build_ctx_in_user_scope(
        self, ctx: ClientContext, scope: UserScope
    ) -> ProjectPermissionContext:
        return ProjectPermissionContext()

    async def build_ctx_in_container_registry_scope(
        self, _ctx: ClientContext, scope: ContainerRegistryScope
    ) -> ProjectPermissionContext:
        permissions = MEMBER_PERMISSIONS
        return ProjectPermissionContext(
            registry_id_to_additional_permission_map={scope.registry_id: permissions}
        )

    @override
    @classmethod
    async def _permission_for_owner(
        cls,
    ) -> frozenset[ProjectPermission]:
        return OWNER_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_admin(
        cls,
    ) -> frozenset[ProjectPermission]:
        return ADMIN_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_monitor(
        cls,
    ) -> frozenset[ProjectPermission]:
        return MONITOR_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_privileged_member(
        cls,
    ) -> frozenset[ProjectPermission]:
        return PRIVILEGED_MEMBER_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_member(
        cls,
    ) -> frozenset[ProjectPermission]:
        return MEMBER_PERMISSIONS


async def get_projects(
    target_scope: ScopeType,
    requested_permission: ProjectPermission,
    project_id: uuid.UUID | None = None,
    project_name: str | None = None,
    *,
    ctx: ClientContext,
    db_conn: SAConnection,
) -> list[ProjectModel]:
    async with ctx.db.begin_readonly_session(db_conn) as db_session:
        builder = ProjectPermissionContextBuilder(db_session)
        permission_ctx = await builder.build(ctx, target_scope, requested_permission)
        query_stmt = await permission_ctx.build_query()
        if query_stmt is None:
            return []
        if project_id is not None:
            query_stmt = query_stmt.where(GroupRow.id == project_id)
        if project_name is not None:
            query_stmt = query_stmt.where(GroupRow.name == project_name)
        result: list[ProjectModel] = []
        async for row in await db_session.stream_scalars(query_stmt):
            permissions = await permission_ctx.calculate_final_permission(row)
            result.append(ProjectModel.from_row(row, permissions))
    return result


async def get_permission_ctx(
    db_conn: SAConnection,
    ctx: ClientContext,
    requested_permission: ProjectPermission,
    target_scope: ScopeType,
    container_registry_scope: ContainerRegistryScope | None = None,
) -> ProjectPermissionContext:
    async with ctx.db.begin_readonly_session(db_conn) as db_session:
        builder = ProjectPermissionContextBuilder(db_session)

        if container_registry_scope is not None:
            return await builder.build_ctx_in_container_registry_scope(
                ctx, container_registry_scope
            )
        return await builder.build(ctx, target_scope, requested_permission)



"""sync_kernelssessionname_to_real_session_name

Revision ID: dddf9be580f5
Revises: 857b763b8618
Create Date: 2024-04-01 16:58:14.341114

"""

import textwrap

from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "dddf9be580f5"
down_revision = "857b763b8618"
branch_labels = None
depends_on = None


def upgrade() -> None:
    conn = op.get_bind()
    sync_stmt = textwrap.dedent(
        """
        UPDATE kernels
        SET session_name = sessions.name
        FROM sessions
        WHERE kernels.session_id = sessions.id
        AND kernels.session_name <> sessions.name;
        """
    )
    conn.execute(text(sync_stmt))


def downgrade() -> None:
    pass



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.app_config.repository import AppConfigRepository
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class AppConfigRepositories:
    repository: AppConfigRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = AppConfigRepository(args.db, args.valkey_stat_client)

        return cls(
            repository=repository,
        )



from .get_all import GetAllNamespacesAction, GetAllNamespacesActionResult
from .get_multi import GetNamespacesAction, GetNamespacesActionResult
from .register import RegisterNamespaceAction, RegisterNamespaceActionResult
from .search import SearchStorageNamespacesAction, SearchStorageNamespacesActionResult
from .unregister import UnregisterNamespaceAction, UnregisterNamespaceActionResult

__all__ = [
    "GetAllNamespacesAction",
    "GetAllNamespacesActionResult",
    "GetNamespacesAction",
    "GetNamespacesActionResult",
    "RegisterNamespaceAction",
    "RegisterNamespaceActionResult",
    "SearchStorageNamespacesAction",
    "SearchStorageNamespacesActionResult",
    "UnregisterNamespaceAction",
    "UnregisterNamespaceActionResult",
]



"""Fair Share fetcher module."""

from __future__ import annotations

from .domain import (
    fetch_domain_fair_shares,
    fetch_rg_domain_fair_shares,
    get_domain_fair_share_pagination_spec,
)
from .project import (
    fetch_project_fair_shares,
    fetch_rg_project_fair_shares,
    get_project_fair_share_pagination_spec,
)
from .user import (
    fetch_rg_user_fair_shares,
    fetch_user_fair_shares,
    get_user_fair_share_pagination_spec,
)

__all__ = [
    # Domain
    "fetch_domain_fair_shares",
    "fetch_rg_domain_fair_shares",
    "get_domain_fair_share_pagination_spec",
    # Project
    "fetch_project_fair_shares",
    "fetch_rg_project_fair_shares",
    "get_project_fair_share_pagination_spec",
    # User
    "fetch_user_fair_shares",
    "fetch_rg_user_fair_shares",
    "get_user_fair_share_pagination_spec",
]



import logging
from collections.abc import AsyncIterator
from typing import override

import aiohttp

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.exceptions import ContainerRegistryProjectEmpty

from .base import (
    BaseContainerRegistry,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class GitHubRegistry(BaseContainerRegistry):
    @override
    async def fetch_repositories(self, sess: aiohttp.ClientSession) -> AsyncIterator[str]:
        if not self.registry_info.project:
            raise ContainerRegistryProjectEmpty(self.registry_info.type, self.registry_info.project)

        project = self.registry_info.project
        access_token = self.registry_info.password
        extra = self.registry_info.extra or {}
        entity_type = extra.get("entity_type", None)

        if entity_type is None:
            raise RuntimeError("Entity type is not provided for GitHub registry!")

        base_url = f"https://api.github.com/{entity_type}/{project}/packages"

        headers = {
            "Authorization": f"Bearer {access_token}",
            "Accept": "application/vnd.github.v3+json",
        }
        page = 1

        while True:
            async with sess.get(
                base_url,
                headers=headers,
                params={"package_type": "container", "per_page": 30, "page": page},
            ) as response:
                if response.status == 200:
                    data = await response.json()
                    for repo in data:
                        yield f"{project}/{repo['name']}"
                    if "next" in response.links:
                        page += 1
                    else:
                        break
                else:
                    raise RuntimeError(
                        f"Failed to fetch repositories! {response.status} error occured."
                    )



"""Query conditions and orders for image repository."""

from __future__ import annotations

import uuid
from collections.abc import Collection

import sqlalchemy as sa

from ai.backend.common.types import ImageID
from ai.backend.manager.api.gql.base import StringMatchSpec
from ai.backend.manager.data.image.types import ImageStatus
from ai.backend.manager.models.image import ImageAliasRow, ImageRow
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder


class ImageConditions:
    """Query conditions for images."""

    @staticmethod
    def by_ids(image_ids: Collection[ImageID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageRow.id.in_(image_ids)

        return inner

    @staticmethod
    def by_canonicals(canonicals: Collection[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageRow.name.in_(canonicals)

        return inner

    @staticmethod
    def by_statuses(statuses: Collection[ImageStatus]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageRow.status.in_(statuses)

        return inner

    @staticmethod
    def by_architecture(architecture: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageRow.architecture == architecture

        return inner

    @staticmethod
    def by_registry_id(registry_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageRow.registry_id == registry_id

        return inner

    # String filter factories for name
    @staticmethod
    def by_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageRow.name.ilike(f"%{spec.value}%")
            else:
                condition = ImageRow.name.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(ImageRow.name) == spec.value.lower()
            else:
                condition = ImageRow.name == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageRow.name.ilike(f"{spec.value}%")
            else:
                condition = ImageRow.name.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageRow.name.ilike(f"%{spec.value}")
            else:
                condition = ImageRow.name.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    # String filter factories for architecture
    @staticmethod
    def by_architecture_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageRow.architecture.ilike(f"%{spec.value}%")
            else:
                condition = ImageRow.architecture.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_architecture_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(ImageRow.architecture) == spec.value.lower()
            else:
                condition = ImageRow.architecture == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_architecture_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageRow.architecture.ilike(f"{spec.value}%")
            else:
                condition = ImageRow.architecture.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_architecture_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageRow.architecture.ilike(f"%{spec.value}")
            else:
                condition = ImageRow.architecture.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner


class ImageOrders:
    """Query orders for images."""

    @staticmethod
    def name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.name.asc()
        return ImageRow.name.desc()

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.created_at.asc()
        return ImageRow.created_at.desc()

    @staticmethod
    def architecture(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.architecture.asc()
        return ImageRow.architecture.desc()

    @staticmethod
    def size_bytes(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.size_bytes.asc()
        return ImageRow.size_bytes.desc()

    @staticmethod
    def registry(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.registry.asc()
        return ImageRow.registry.desc()

    @staticmethod
    def tag(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.tag.asc()
        return ImageRow.tag.desc()

    @staticmethod
    def status(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.status.asc()
        return ImageRow.status.desc()


class ImageAliasConditions:
    """Query conditions for image aliases."""

    @staticmethod
    def by_ids(alias_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageAliasRow.id.in_(alias_ids)

        return inner

    @staticmethod
    def by_image_ids(image_ids: Collection[ImageID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageAliasRow.image_id.in_(image_ids)

        return inner

    # String filter factories for alias
    @staticmethod
    def by_alias_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageAliasRow.alias.ilike(f"%{spec.value}%")
            else:
                condition = ImageAliasRow.alias.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_alias_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(ImageAliasRow.alias) == spec.value.lower()
            else:
                condition = ImageAliasRow.alias == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_alias_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageAliasRow.alias.ilike(f"{spec.value}%")
            else:
                condition = ImageAliasRow.alias.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_alias_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageAliasRow.alias.ilike(f"%{spec.value}")
            else:
                condition = ImageAliasRow.alias.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner


class ImageAliasOrders:
    """Query orders for image aliases."""

    @staticmethod
    def alias(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageAliasRow.alias.asc()
        return ImageAliasRow.alias.desc()



from pathlib import PurePosixPath

from ai.backend.common.dto.storage.request import FileDeleteAsyncRequest
from ai.backend.common.types import (
    VFolderHostPermission,
    VFolderID,
)
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.errors.storage import VFolderInvalidParameter
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.vfolder import (
    is_unmanaged,
)
from ai.backend.manager.repositories.user.repository import UserRepository
from ai.backend.manager.repositories.vfolder.repository import VfolderRepository
from ai.backend.manager.services.vfolder.actions.file import (
    CreateDownloadSessionAction,
    CreateDownloadSessionActionResult,
    CreateUploadSessionAction,
    CreateUploadSessionActionResult,
    DeleteFilesAction,
    DeleteFilesActionResult,
    DeleteFilesAsyncAction,
    DeleteFilesAsyncActionResult,
    ListFilesAction,
    ListFilesActionResult,
    MkdirAction,
    MkdirActionResult,
    RenameFileAction,
    RenameFileActionResult,
)
from ai.backend.manager.services.vfolder.types import FileInfo


class VFolderFileService:
    _config_provider: ManagerConfigProvider
    _storage_manager: StorageSessionManager
    _vfolder_repository: VfolderRepository
    _user_repository: UserRepository

    def __init__(
        self,
        config_provider: ManagerConfigProvider,
        storage_manager: StorageSessionManager,
        vfolder_repository: VfolderRepository,
        user_repository: UserRepository,
    ) -> None:
        self._config_provider = config_provider
        self._storage_manager = storage_manager
        self._vfolder_repository = vfolder_repository
        self._user_repository = user_repository

    async def upload_file(
        self, action: CreateUploadSessionAction
    ) -> CreateUploadSessionActionResult:
        # Get VFolder data using repository
        user = await self._user_repository.get_user_by_uuid(action.user_uuid)
        if not user.domain_name:
            raise VFolderInvalidParameter("User has no domain assigned")
        vfolder_data = await self._vfolder_repository.get_by_id_validated(
            action.vfolder_uuid, user.id, user.domain_name
        )
        if not vfolder_data:
            raise VFolderInvalidParameter("VFolder not found")

        # Check host permissions
        allowed_vfolder_types = (
            await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
        )
        await self._vfolder_repository.ensure_host_permission_allowed(
            vfolder_data.host,
            permission=VFolderHostPermission.UPLOAD_FILE,
            allowed_vfolder_types=allowed_vfolder_types,
            user_uuid=action.user_uuid,
            resource_policy=action.keypair_resource_policy,
            domain_name=vfolder_data.domain_name,
        )

        proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(
            vfolder_data.host, is_unmanaged(vfolder_data.unmanaged_path)
        )

        # Create VFolderID from data
        vfolder_id = VFolderID(
            quota_scope_id=vfolder_data.quota_scope_id,
            folder_id=vfolder_data.id,
        )

        manager_client = self._storage_manager.get_manager_facing_client(proxy_name)
        storage_reply = await manager_client.upload_file(
            volume_name,
            str(vfolder_id),
            action.path,
            action.size,
        )
        client_api_url = self._storage_manager.get_client_api_url(proxy_name)
        return CreateUploadSessionActionResult(
            vfolder_uuid=action.vfolder_uuid,
            token=storage_reply["token"],
            url=str(client_api_url / "upload"),
        )

    async def download_file(
        self, action: CreateDownloadSessionAction
    ) -> CreateDownloadSessionActionResult:
        # Get VFolder data using repository
        user = await self._user_repository.get_user_by_uuid(action.user_uuid)
        if not user.domain_name:
            raise VFolderInvalidParameter("User has no domain assigned")
        vfolder_data = await self._vfolder_repository.get_by_id_validated(
            action.vfolder_uuid, user.id, user.domain_name
        )
        if not vfolder_data:
            raise VFolderInvalidParameter("VFolder not found")

        # Check host permissions
        allowed_vfolder_types = (
            await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
        )
        await self._vfolder_repository.ensure_host_permission_allowed(
            vfolder_data.host,
            permission=VFolderHostPermission.DOWNLOAD_FILE,
            allowed_vfolder_types=allowed_vfolder_types,
            user_uuid=action.user_uuid,
            resource_policy=action.keypair_resource_policy,
            domain_name=vfolder_data.domain_name,
        )

        proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(
            vfolder_data.host, is_unmanaged(vfolder_data.unmanaged_path)
        )

        # Create VFolderID from data
        vfolder_id = VFolderID(
            quota_scope_id=vfolder_data.quota_scope_id,
            folder_id=vfolder_data.id,
        )

        manager_client = self._storage_manager.get_manager_facing_client(proxy_name)
        # For download, we need to handle the request differently as it includes extra params
        storage_reply = await manager_client.download_file(
            volume=volume_name,
            vfid=str(vfolder_id),
            relpath=action.path,
            archive=action.archive,
            unmanaged_path=vfolder_data.unmanaged_path,
        )
        client_api_url = self._storage_manager.get_client_api_url(proxy_name)
        return CreateDownloadSessionActionResult(
            vfolder_uuid=action.vfolder_uuid,
            token=storage_reply["token"],
            url=str(client_api_url / "download"),
        )

    async def list_files(self, action: ListFilesAction) -> ListFilesActionResult:
        # Get user info and check VFolder access using repository
        user = await self._user_repository.get_user_by_uuid(action.user_uuid)
        if not user.domain_name:
            raise VFolderInvalidParameter("User has no domain assigned")
        vfolder_data = await self._vfolder_repository.get_by_id_validated(
            action.vfolder_uuid, user.id, user.domain_name
        )

        proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(
            vfolder_data.host, is_unmanaged(vfolder_data.unmanaged_path)
        )

        # Create VFolderID from data
        vfolder_id = VFolderID(
            quota_scope_id=vfolder_data.quota_scope_id,
            folder_id=vfolder_data.id,
        )

        manager_client = self._storage_manager.get_manager_facing_client(proxy_name)
        result = await manager_client.list_files(
            volume_name,
            str(vfolder_id),
            action.path,
        )
        return ListFilesActionResult(
            vfolder_uuid=action.vfolder_uuid,
            files=[
                FileInfo(
                    name=item["name"],
                    type=item["type"],
                    size=item["stat"]["size"],
                    mode=item["stat"]["mode"],
                    created=item["stat"]["created"],
                    modified=item["stat"]["modified"],
                )
                for item in result["items"]
            ],
        )

    async def rename_file(self, action: RenameFileAction) -> RenameFileActionResult:
        # Get VFolder data using repository
        user = await self._user_repository.get_user_by_uuid(action.user_uuid)
        if not user.domain_name:
            raise VFolderInvalidParameter("User has no domain assigned")
        vfolder_data = await self._vfolder_repository.get_by_id_validated(
            action.vfolder_uuid, user.id, user.domain_name
        )
        if not vfolder_data:
            raise VFolderInvalidParameter("VFolder not found")

        # Check host permissions
        allowed_vfolder_types = (
            await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
        )
        await self._vfolder_repository.ensure_host_permission_allowed(
            vfolder_data.host,
            permission=VFolderHostPermission.MODIFY,
            allowed_vfolder_types=allowed_vfolder_types,
            user_uuid=action.user_uuid,
            resource_policy=action.keypair_resource_policy,
            domain_name=vfolder_data.domain_name,
        )

        proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(
            vfolder_data.host, is_unmanaged(vfolder_data.unmanaged_path)
        )

        # Create VFolderID from data
        vfolder_id = VFolderID(
            quota_scope_id=vfolder_data.quota_scope_id,
            folder_id=vfolder_data.id,
        )

        manager_client = self._storage_manager.get_manager_facing_client(proxy_name)
        await manager_client.rename_file(
            volume_name,
            str(vfolder_id),
            action.target_path,
            action.new_name,
        )
        return RenameFileActionResult(vfolder_uuid=action.vfolder_uuid)

    async def delete_files(self, action: DeleteFilesAction) -> DeleteFilesActionResult:
        # Get user info and check VFolder access using repository
        user = await self._user_repository.get_user_by_uuid(action.user_uuid)
        if not user.domain_name:
            raise VFolderInvalidParameter("User has no domain assigned")
        vfolder_data = await self._vfolder_repository.get_by_id_validated(
            action.vfolder_uuid, user.id, user.domain_name
        )

        if not vfolder_data:
            raise VFolderInvalidParameter("The specified vfolder is not accessible.")

        proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(
            vfolder_data.host, is_unmanaged(vfolder_data.unmanaged_path)
        )

        # Create VFolderID from data
        vfolder_id = VFolderID(
            quota_scope_id=vfolder_data.quota_scope_id,
            folder_id=vfolder_data.id,
        )

        manager_client = self._storage_manager.get_manager_facing_client(proxy_name)
        await manager_client.delete_files(
            volume_name,
            str(vfolder_id),
            action.files,
            action.recursive,
        )
        return DeleteFilesActionResult(vfolder_uuid=action.vfolder_uuid)

    async def delete_files_async(
        self, action: DeleteFilesAsyncAction
    ) -> DeleteFilesAsyncActionResult:
        # Get user info and check VFolder access using repository
        user = await self._user_repository.get_user_by_uuid(action.user_uuid)
        if not user.domain_name:
            raise VFolderInvalidParameter("User has no domain assigned")
        vfolder_data = await self._vfolder_repository.get_by_id_validated(
            action.vfolder_uuid, user.id, user.domain_name
        )

        if not vfolder_data:
            raise VFolderInvalidParameter("The specified vfolder is not accessible.")

        proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(
            vfolder_data.host, is_unmanaged(vfolder_data.unmanaged_path)
        )

        # Create VFolderID from data
        vfolder_id = VFolderID(
            quota_scope_id=vfolder_data.quota_scope_id,
            folder_id=vfolder_data.id,
        )

        # Create request DTO
        request = FileDeleteAsyncRequest(
            volume=volume_name,
            vfid=vfolder_id,
            relpaths=[PurePosixPath(file) for file in action.files],
            recursive=action.recursive,
        )

        # Call storage proxy client
        manager_client = self._storage_manager.get_manager_facing_client(proxy_name)
        response = await manager_client.delete_files_async(request)

        return DeleteFilesAsyncActionResult(
            vfolder_uuid=action.vfolder_uuid, task_id=response.bgtask_id
        )

    async def mkdir(self, action: MkdirAction) -> MkdirActionResult:
        if isinstance(action.path, list) and len(action.path) > 50:
            raise VFolderInvalidParameter("Too many directories specified.")

        # Get VFolder data using repository
        user = await self._user_repository.get_user_by_uuid(action.user_id)
        if not user.domain_name:
            raise VFolderInvalidParameter("User has no domain assigned")
        vfolder_data = await self._vfolder_repository.get_by_id_validated(
            action.vfolder_uuid, user.id, user.domain_name
        )
        if not vfolder_data:
            raise VFolderInvalidParameter("VFolder not found")

        proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(
            vfolder_data.host, is_unmanaged(vfolder_data.unmanaged_path)
        )

        # Create VFolderID from data
        vfolder_id = VFolderID(
            quota_scope_id=vfolder_data.quota_scope_id,
            folder_id=vfolder_data.id,
        )

        manager_client = self._storage_manager.get_manager_facing_client(proxy_name)
        storage_reply = await manager_client.mkdir(
            volume=volume_name,
            vfid=str(vfolder_id),
            relpath=action.path,
            exist_ok=action.exist_ok,
            parents=action.parents,
        )
        results = storage_reply["results"]
        return MkdirActionResult(
            vfolder_uuid=action.vfolder_uuid,
            results=results,
            storage_resp_status=200,
        )



"""Storage source implementation for deployment repository."""

import tomli

from ai.backend.common.types import VFolderID
from ai.backend.manager.data.deployment.types import ModelServiceDefinition
from ai.backend.manager.data.vfolder.types import VFolderLocation
from ai.backend.manager.errors.deployment import DefinitionFileNotFound
from ai.backend.manager.models.storage import StorageSessionManager


class DeploymentStorageSource:
    """Storage source for deployment-related file operations."""

    _storage_manager: StorageSessionManager

    def __init__(self, storage_manager: StorageSessionManager) -> None:
        self._storage_manager = storage_manager

    async def fetch_service_config(
        self,
        model_vfolder: VFolderLocation,
    ) -> ModelServiceDefinition | None:
        """
        Fetch and parse service-definition.toml from model vfolder.

        Args:
            model_vfolder: The model vfolder location information

        Returns:
            Parsed service definition as ModelServiceDefinition, or None if not found
        """
        try:
            vfid = VFolderID(model_vfolder.quota_scope_id, model_vfolder.id)
            folder_host = model_vfolder.host

            proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(folder_host)
            manager_client = self._storage_manager.get_manager_facing_client(proxy_name)

            chunks = await manager_client.fetch_file_content(
                volume_name,
                str(vfid),
                "./service-definition.toml",
            )

            if chunks:
                raw_content = chunks.decode("utf-8")
                parsed_toml = tomli.loads(raw_content)
                return ModelServiceDefinition(**parsed_toml)

        except Exception:
            # Service definition file not found or parse error
            pass

        return None

    async def fetch_definition_file(
        self,
        vfolder_location: VFolderLocation,
        definition_candidates: list[str],
    ) -> bytes:
        """
        Fetch definition file from model vfolder.
        Args:
            vfolder_location: The model vfolder location information
            definition_candidates: List of candidate file paths to look for
        """
        vfid = VFolderID(vfolder_location.quota_scope_id, vfolder_location.id)
        folder_host = vfolder_location.host

        proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(folder_host)
        manager_client = self._storage_manager.get_manager_facing_client(proxy_name)

        for definition_file_path in definition_candidates:
            try:
                # Try to fetch the file to check if it exists
                file_content = await manager_client.fetch_file_content(
                    volume_name,
                    str(vfid),
                    f"{definition_file_path}",
                )
                if file_content:
                    return file_content
            except Exception:
                continue
        raise DefinitionFileNotFound(
            f"definition file not found in vfolder {vfid} for candidates {definition_candidates}"
        )



"""add_resource_slot_normalization_tables

Revision ID: 4b7b4b040a70
Revises: bc4e0e948300
Create Date: 2026-02-08 12:00:00.000000

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "4b7b4b040a70"
down_revision = "bc4e0e948300"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # 1. resource_slot_types
    op.create_table(
        "resource_slot_types",
        sa.Column("slot_name", sa.String(length=64), nullable=False),
        sa.Column("slot_type", sa.String(length=16), nullable=False),
        sa.Column("display_name", sa.String(length=128), nullable=True),
        sa.Column("rank", sa.Integer(), nullable=False, server_default=sa.text("0")),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("slot_name", name=op.f("pk_resource_slot_types")),
    )

    # 2. agent_resources
    op.create_table(
        "agent_resources",
        sa.Column("agent_id", sa.String(length=64), nullable=False),
        sa.Column("slot_name", sa.String(length=64), nullable=False),
        sa.Column("capacity", sa.Numeric(precision=24, scale=6), nullable=False),
        sa.Column("used", sa.Numeric(precision=24, scale=6), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("agent_id", "slot_name", name=op.f("pk_agent_resources")),
        sa.ForeignKeyConstraint(
            ["agent_id"],
            ["agents.id"],
            name=op.f("fk_agent_resources_agent_id_agents"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["slot_name"],
            ["resource_slot_types.slot_name"],
            name=op.f("fk_agent_resources_slot_name_resource_slot_types"),
        ),
    )
    op.create_index(
        op.f("ix_agent_resources_slot_name"),
        "agent_resources",
        ["slot_name"],
        unique=False,
    )
    op.create_index(
        "ix_agent_resources_agent_avail",
        "agent_resources",
        ["agent_id", "slot_name", "capacity", "used"],
        unique=False,
    )

    # 3. resource_allocations
    op.create_table(
        "resource_allocations",
        sa.Column("kernel_id", GUID(), nullable=False),
        sa.Column("slot_name", sa.String(length=64), nullable=False),
        sa.Column("requested", sa.Numeric(precision=24, scale=6), nullable=False),
        sa.Column("used", sa.Numeric(precision=24, scale=6), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column("used_at", sa.DateTime(timezone=True), nullable=True),
        sa.PrimaryKeyConstraint("kernel_id", "slot_name", name=op.f("pk_resource_allocations")),
        sa.ForeignKeyConstraint(
            ["kernel_id"],
            ["kernels.id"],
            name=op.f("fk_resource_allocations_kernel_id_kernels"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["slot_name"],
            ["resource_slot_types.slot_name"],
            name=op.f("fk_resource_allocations_slot_name_resource_slot_types"),
        ),
    )
    op.create_index(
        op.f("ix_resource_allocations_slot_name"),
        "resource_allocations",
        ["slot_name"],
        unique=False,
    )
    op.create_index(
        "ix_ra_kernel_slot",
        "resource_allocations",
        ["kernel_id", "slot_name"],
        unique=False,
    )

    # 4. Seed data for resource_slot_types
    conn = op.get_bind()
    conn.execute(
        sa.text("""
            INSERT INTO resource_slot_types
                (slot_name, slot_type, display_name, rank)
            VALUES
                ('cpu', 'count', 'CPU', 40),
                ('mem', 'bytes', 'Memory', 50),
                ('cuda.device', 'count', 'GPU (CUDA)', 10),
                ('cuda.shares', 'count', 'GPU (fGPU)', 20),
                ('rocm.device', 'count', 'GPU (ROCm)', 30),
                ('tpu.device', 'count', 'TPU', 35)
        """)
    )

    # 5. Backfill: register any unknown slot names from agents and kernels
    conn.execute(
        sa.text("""
            INSERT INTO resource_slot_types (slot_name, slot_type, rank)
            SELECT DISTINCT kv.key, 'count', 0
            FROM agents a, jsonb_each_text(a.available_slots) AS kv(key, value)
            WHERE a.status = 'ALIVE'
              AND kv.key NOT IN (SELECT slot_name FROM resource_slot_types)
            ON CONFLICT (slot_name) DO NOTHING
        """)
    )
    conn.execute(
        sa.text("""
            INSERT INTO resource_slot_types (slot_name, slot_type, rank)
            SELECT DISTINCT req.key, 'count', 0
            FROM kernels k, jsonb_each_text(k.requested_slots) AS req(key, value)
            WHERE k.status NOT IN ('TERMINATED', 'CANCELLED')
              AND req.key NOT IN (SELECT slot_name FROM resource_slot_types)
            ON CONFLICT (slot_name) DO NOTHING
        """)
    )

    # 6. Backfill agent_resources from agents JSONB columns
    conn.execute(
        sa.text("""
            INSERT INTO agent_resources
                (agent_id, slot_name, capacity, used)
            SELECT a.id, kv.key, kv.value::numeric(24,6), occ.value::numeric(24,6)
            FROM agents a,
                 jsonb_each_text(a.available_slots) AS kv(key, value)
            LEFT JOIN jsonb_each_text(a.occupied_slots) AS occ(key, value)
                 ON kv.key = occ.key
            WHERE a.status = 'ALIVE'
              AND kv.key IN (SELECT slot_name FROM resource_slot_types)
            ON CONFLICT (agent_id, slot_name) DO NOTHING
        """)
    )

    # 7. Backfill resource_allocations from kernels JSONB columns
    conn.execute(
        sa.text("""
            INSERT INTO resource_allocations
                (kernel_id, slot_name, requested, used)
            SELECT k.id, req.key, req.value::numeric(24,6), occ.value::numeric(24,6)
            FROM kernels k,
                 jsonb_each_text(k.requested_slots) AS req(key, value)
            LEFT JOIN jsonb_each_text(k.occupied_slots) AS occ(key, value)
                 ON req.key = occ.key
            WHERE k.status NOT IN ('TERMINATED', 'CANCELLED')
              AND req.key IN (SELECT slot_name FROM resource_slot_types)
            ON CONFLICT (kernel_id, slot_name) DO NOTHING
        """)
    )


def downgrade() -> None:
    op.drop_index("ix_ra_kernel_slot", table_name="resource_allocations")
    op.drop_index(op.f("ix_resource_allocations_slot_name"), table_name="resource_allocations")
    op.drop_table("resource_allocations")
    op.drop_index("ix_agent_resources_agent_avail", table_name="agent_resources")
    op.drop_index(
        op.f("ix_agent_resources_slot_name"),
        table_name="agent_resources",
    )
    op.drop_table("agent_resources")
    op.drop_table("resource_slot_types")



"""add-max-session-lifetime

Revision ID: 81c264528f20
Revises: d727b5da20e6
Create Date: 2022-04-21 09:22:01.405710

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "81c264528f20"
down_revision = "d727b5da20e6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        "keypair_resource_policies",
        sa.Column(
            "max_session_lifetime", sa.Integer(), server_default=sa.text("0"), nullable=False
        ),
    )


def downgrade() -> None:
    op.drop_column("keypair_resource_policies", "max_session_lifetime")



from __future__ import annotations

import enum
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Protocol,
    TypeVar,
)

import attr
from graphql import UndefinedType
from pydantic import AliasChoices, BaseModel, Field
from strawberry.types.unset import UnsetType

from ai.backend.common.types import MountPermission, MountTypes

if TYPE_CHECKING:
    from ai.backend.common.lock import AbstractDistributedLock

    from .defs import LockID


# Sentinel is a special object that indicates a special status instead of a value
# where the user expects a value.
# According to the discussion in https://github.com/python/typing/issues/236,
# we define our Sentinel type as an enum with only one special value.
# This enables passing of type checks by "value is sentinel" (or "value is Sentinel.token")
# instead of more expensive "isinstance(value, Sentinel)" because we can assure type checkers
# to think there is no other possible instances of the Sentinel type.


class Sentinel(enum.Enum):
    TOKEN = 0


_SENTINEL = Sentinel.TOKEN


@attr.define(slots=True)
class UserScope:
    domain_name: str
    group_id: uuid.UUID
    user_uuid: uuid.UUID
    user_role: str


class DistributedLockFactory(Protocol):
    def __call__(self, lock_id: LockID, lifetime_hint: float) -> AbstractDistributedLock: ...


class MountOptionModel(BaseModel):
    mount_destination: Annotated[
        str | None,
        Field(description="Mount destination, defaults to /home/work/{folder_name}.", default=None),
    ]
    type: Annotated[MountTypes, Field(default=MountTypes.BIND)]
    permission: Annotated[
        MountPermission | None,
        Field(validation_alias=AliasChoices("permission", "perm"), default=None),
    ]


class PartialModifier(ABC):
    @abstractmethod
    def fields_to_update(self) -> dict[str, Any]:
        """
        Returns a dictionary of fields that should be updated.
        This is different from to_dict() as it specifically maps fields to their storage keys.
        """
        pass


class _TriStateEnum(enum.Enum):
    UPDATE = "update"
    NULLIFY = "nullify"
    NOP = "nop"


TVal = TypeVar("TVal")


@dataclass
class TriState[TVal]:
    """
    TriState is a class that represents partial updates to an attribute of an object.
    It is used to indicate whether an attribute should be updated, set to None, or not modified at all.
    It can be in one of three states:
    - UPDATE: The attribute should be updated with the given value.
    - NULLIFY: The attribute should be set to None.
    - NOP: No operation should be performed on the attribute.
    """

    _state: _TriStateEnum
    _value: TVal | None

    def __init__(self, state: _TriStateEnum, value: TVal | None) -> None:
        """
        Initialize a TriState object with the given state and value.
        Do not call this constructor directly. Use the class methods instead.
        """
        self._state = state
        self._value = value

    @classmethod
    def from_graphql(cls, value: TVal | None | UndefinedType) -> TriState[TVal]:
        if value is None:
            return cls.nullify()
        if isinstance(value, (UndefinedType, UnsetType)):
            return cls.nop()
        return cls.update(value)

    @classmethod
    def update(cls, value: TVal) -> TriState[TVal]:
        return cls(state=_TriStateEnum.UPDATE, value=value)

    @classmethod
    def nullify(cls) -> TriState[TVal]:
        return cls(state=_TriStateEnum.NULLIFY, value=None)

    @classmethod
    def nop(cls) -> TriState[TVal]:
        return cls(state=_TriStateEnum.NOP, value=None)

    def value(self) -> TVal:
        """
        Returns the value of the TriState object.
        It should only be used when the state value is unambiguously UPDATE.
        """
        if self._state != _TriStateEnum.UPDATE:
            raise ValueError("Not allowed to get value when state is not UPDATE")
        if self._value is None:
            raise ValueError("TriState value is not set when state is UPDATE")
        return self._value

    def optional_value(self) -> TVal | None:
        """
        Returns the value of the TriState object.
        When state is not UPDATE, it returns None.
        This is useful for cases where you want to check if the state is UPDATE
        and get the value, or if it is NULLIFY or NOP and get None.
        """
        if self._state == _TriStateEnum.UPDATE:
            return self._value
        return None

    def update_dict(self, dict: dict[str, Any], attr_name: str) -> None:
        match self._state:
            case _TriStateEnum.UPDATE:
                dict[attr_name] = self._value
            case _TriStateEnum.NULLIFY:
                dict[attr_name] = None
            case _TriStateEnum.NOP:
                pass


class OptionalState[TVal]:
    """
    OptionalState is a class that represents partial updates to an attribute of an object.
    It is used to indicate whether an attribute should be updated or not modified at all.
    It can be in one of two states:
    - UPDATE: The attribute should be updated with the given value.
    - NOP: No operation should be performed on the attribute.
    This class is similar to TriState, but it cannot be in the NULLIFY state.
    """

    _state: _TriStateEnum
    _value: TVal | None

    def __init__(self, state: _TriStateEnum, value: TVal | None) -> None:
        if state == _TriStateEnum.NULLIFY:
            raise ValueError("OptionalState cannot be NULLIFY")
        self._state = state
        self._value = value

    @classmethod
    def from_graphql(cls, value: TVal | None | UndefinedType | UnsetType) -> OptionalState[TVal]:
        if isinstance(value, (UndefinedType, UnsetType)):
            return OptionalState.nop()
        if value is None:
            raise ValueError("OptionalState cannot be NULLIFY")
        return OptionalState.update(value)

    @classmethod
    def update(cls, value: TVal) -> OptionalState[TVal]:
        return cls(state=_TriStateEnum.UPDATE, value=value)

    @classmethod
    def nop(cls) -> OptionalState[TVal]:
        return cls(state=_TriStateEnum.NOP, value=None)

    def value(self) -> TVal:
        """
        Returns the value of the TriState object.
        It should only be used when the state value is unambiguously UPDATE.
        """
        if self._state != _TriStateEnum.UPDATE:
            raise ValueError("Not allowed to get value when state is not UPDATE")
        if self._value is None:
            raise ValueError("TriState value is not set when state is UPDATE")
        return self._value

    def optional_value(self) -> TVal | None:
        """
        Returns the value of the TriState object.
        When state is not UPDATE, it returns None.
        This is useful for cases where you want to check if the state is UPDATE
        and get the value, or if it is NULLIFY or NOP and get None.
        """
        if self._state == _TriStateEnum.UPDATE:
            return self._value
        return None

    def update_dict(self, dict: dict[str, Any], attr_name: str) -> None:
        match self._state:
            case _TriStateEnum.UPDATE:
                dict[attr_name] = self._value
            case _TriStateEnum.NOP:
                pass


class SMTPTriggerPolicy(enum.StrEnum):
    ALL = "ALL"
    ON_ERROR = "ON_ERROR"


@dataclass
class OffsetBasedPaginationOptions:
    """Standard offset/limit pagination options."""

    offset: int | None = None
    limit: int | None = None


@dataclass
class ForwardPaginationOptions:
    """Forward pagination: fetch items after a given cursor."""

    after: str | None = None
    first: int | None = None


@dataclass
class BackwardPaginationOptions:
    """Backward pagination: fetch items before a given cursor."""

    before: str | None = None
    last: int | None = None


@dataclass
class PaginationOptions:
    forward: ForwardPaginationOptions | None = None
    backward: BackwardPaginationOptions | None = None
    offset: OffsetBasedPaginationOptions | None = None



from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.common.types import RuleId
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.services.model_serving.actions.base import ModelServiceAction


@dataclass
class DeleteEndpointAutoScalingRuleAction(ModelServiceAction):
    id: RuleId

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.DEPLOYMENT_AUTO_SCALING_RULE

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE


@dataclass
class DeleteEndpointAutoScalingRuleActionResult(BaseActionResult):
    success: bool

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

import logging
from decimal import Decimal

from ai.backend.common.types import ResourceSlot
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.errors.fair_share import InvalidResourceWeightError
from ai.backend.manager.models.scaling_group import ScalingGroupRow
from ai.backend.manager.models.scaling_group.types import FairShareScalingGroupSpec
from ai.backend.manager.repositories.base import Updater
from ai.backend.manager.repositories.scaling_group import ScalingGroupRepository
from ai.backend.manager.repositories.scaling_group.updaters import (
    ResourceGroupFairShareUpdaterSpec,
    ScalingGroupUpdaterSpec,
)
from ai.backend.manager.services.scaling_group.actions.associate_with_domain import (
    AssociateScalingGroupWithDomainsAction,
    AssociateScalingGroupWithDomainsActionResult,
)
from ai.backend.manager.services.scaling_group.actions.associate_with_keypair import (
    AssociateScalingGroupWithKeypairsAction,
    AssociateScalingGroupWithKeypairsActionResult,
)
from ai.backend.manager.services.scaling_group.actions.associate_with_user_group import (
    AssociateScalingGroupWithUserGroupsAction,
    AssociateScalingGroupWithUserGroupsActionResult,
)
from ai.backend.manager.services.scaling_group.actions.create import (
    CreateScalingGroupAction,
    CreateScalingGroupActionResult,
)
from ai.backend.manager.services.scaling_group.actions.disassociate_with_domain import (
    DisassociateScalingGroupWithDomainsAction,
    DisassociateScalingGroupWithDomainsActionResult,
)
from ai.backend.manager.services.scaling_group.actions.disassociate_with_keypair import (
    DisassociateScalingGroupWithKeypairsAction,
    DisassociateScalingGroupWithKeypairsActionResult,
)
from ai.backend.manager.services.scaling_group.actions.disassociate_with_user_group import (
    DisassociateScalingGroupWithUserGroupsAction,
    DisassociateScalingGroupWithUserGroupsActionResult,
)
from ai.backend.manager.services.scaling_group.actions.get_resource_info import (
    GetResourceInfoAction,
    GetResourceInfoActionResult,
)
from ai.backend.manager.services.scaling_group.actions.list_scaling_groups import (
    SearchScalingGroupsAction,
    SearchScalingGroupsActionResult,
)
from ai.backend.manager.services.scaling_group.actions.modify import (
    ModifyScalingGroupAction,
    ModifyScalingGroupActionResult,
)
from ai.backend.manager.services.scaling_group.actions.purge_scaling_group import (
    PurgeScalingGroupAction,
    PurgeScalingGroupActionResult,
)
from ai.backend.manager.services.scaling_group.actions.update_fair_share_spec import (
    UpdateFairShareSpecAction,
    UpdateFairShareSpecActionResult,
)
from ai.backend.manager.types import TriState

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ScalingGroupService:
    _repository: ScalingGroupRepository

    def __init__(self, repository: ScalingGroupRepository) -> None:
        self._repository = repository

    async def search_scaling_groups(
        self, action: SearchScalingGroupsAction
    ) -> SearchScalingGroupsActionResult:
        """Searches scaling groups."""
        result = await self._repository.search_scaling_groups(
            querier=action.querier,
        )

        return SearchScalingGroupsActionResult(
            scaling_groups=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def create_scaling_group(
        self, action: CreateScalingGroupAction
    ) -> CreateScalingGroupActionResult:
        """Creates a scaling group."""
        scaling_group_data = await self._repository.create_scaling_group(action.creator)
        return CreateScalingGroupActionResult(scaling_group=scaling_group_data)

    async def purge_scaling_group(
        self, action: PurgeScalingGroupAction
    ) -> PurgeScalingGroupActionResult:
        """Purges a scaling group and all related sessions and routes."""
        data = await self._repository.purge_scaling_group(action.purger)
        return PurgeScalingGroupActionResult(data=data)

    async def modify_scaling_group(
        self, action: ModifyScalingGroupAction
    ) -> ModifyScalingGroupActionResult:
        """Modifies a scaling group."""
        scaling_group_data = await self._repository.update_scaling_group(action.updater)
        return ModifyScalingGroupActionResult(scaling_group=scaling_group_data)

    async def associate_scaling_group_with_domains(
        self, action: AssociateScalingGroupWithDomainsAction
    ) -> AssociateScalingGroupWithDomainsActionResult:
        """Associates a scaling group with multiple domains."""
        await self._repository.associate_scaling_group_with_domains(action.bulk_creator)
        return AssociateScalingGroupWithDomainsActionResult()

    async def disassociate_scaling_group_with_domains(
        self, action: DisassociateScalingGroupWithDomainsAction
    ) -> DisassociateScalingGroupWithDomainsActionResult:
        """Disassociates a scaling group from multiple domains."""
        await self._repository.disassociate_scaling_group_with_domains(action.purger)
        return DisassociateScalingGroupWithDomainsActionResult()

    async def associate_scaling_group_with_keypairs(
        self, action: AssociateScalingGroupWithKeypairsAction
    ) -> AssociateScalingGroupWithKeypairsActionResult:
        """Associates a scaling group with multiple keypairs."""
        await self._repository.associate_scaling_group_with_keypairs(action.bulk_creator)
        return AssociateScalingGroupWithKeypairsActionResult()

    async def disassociate_scaling_group_with_keypairs(
        self, action: DisassociateScalingGroupWithKeypairsAction
    ) -> DisassociateScalingGroupWithKeypairsActionResult:
        """Disassociates a scaling group from multiple keypairs."""
        await self._repository.disassociate_scaling_group_with_keypairs(action.purger)
        return DisassociateScalingGroupWithKeypairsActionResult()

    async def associate_scaling_group_with_user_groups(
        self, action: AssociateScalingGroupWithUserGroupsAction
    ) -> AssociateScalingGroupWithUserGroupsActionResult:
        """Associates a scaling group with multiple user groups (projects)."""
        await self._repository.associate_scaling_group_with_user_groups(action.bulk_creator)
        return AssociateScalingGroupWithUserGroupsActionResult()

    async def disassociate_scaling_group_with_user_groups(
        self, action: DisassociateScalingGroupWithUserGroupsAction
    ) -> DisassociateScalingGroupWithUserGroupsActionResult:
        """Disassociates a single scaling group from a user group (project)."""
        await self._repository.disassociate_scaling_group_with_user_groups(action.purger)
        return DisassociateScalingGroupWithUserGroupsActionResult()

    async def get_resource_info(self, action: GetResourceInfoAction) -> GetResourceInfoActionResult:
        """Get aggregated resource information for a scaling group.

        Raises:
            ScalingGroupNotFound: If the scaling group does not exist.
        """
        resource_info = await self._repository.get_resource_info(action.scaling_group)
        return GetResourceInfoActionResult(resource_info=resource_info)

    async def update_fair_share_spec(
        self, action: UpdateFairShareSpecAction
    ) -> UpdateFairShareSpecActionResult:
        """Update fair share spec for a scaling group.

        Supports partial updates with resource weight validation and filtering.

        Validation: Input resource_weights must contain only resource types
        that exist in the scaling group's capacity.

        Filtering: After merging, resource_weights are filtered to only include
        types that exist in capacity (to remove stale resource types).

        Raises:
            ScalingGroupNotFound: If the scaling group does not exist.
            InvalidResourceWeightError: If input contains invalid resource types.
        """
        # 1. Get existing scaling group (raises ScalingGroupNotFound if not found)
        existing_sg = await self._repository.get_scaling_group_by_name(action.resource_group)
        existing_spec = existing_sg.fair_share_spec or FairShareScalingGroupSpec()

        # 2. Get ResourceInfo for capacity
        resource_info = await self._repository.get_resource_info(action.resource_group)
        capacity_keys = set(resource_info.capacity.data.keys())

        # 3. Validate: input resource_weights with non-None weight must exist in capacity
        if action.resource_weights:
            # Only validate types that are being set (not deleted)
            input_types = {
                rw.resource_type for rw in action.resource_weights if rw.weight is not None
            }
            invalid_types = input_types - capacity_keys
            if invalid_types:
                raise InvalidResourceWeightError(sorted(invalid_types))

        # 4. Merge: partial input with existing fair_share_spec
        merged_resource_weights = dict(existing_spec.resource_weights.data)
        if action.resource_weights:
            for rw in action.resource_weights:
                if rw.weight is None:
                    # None means delete
                    merged_resource_weights.pop(rw.resource_type, None)
                else:
                    merged_resource_weights[rw.resource_type] = rw.weight

        # 5. Filter: keep only resource types in capacity
        filtered_resource_weights = {
            k: v for k, v in merged_resource_weights.items() if k in capacity_keys
        }

        # 6. Build new spec with merged values
        new_spec = FairShareScalingGroupSpec(
            half_life_days=action.half_life_days
            if action.half_life_days is not None
            else existing_spec.half_life_days,
            lookback_days=action.lookback_days
            if action.lookback_days is not None
            else existing_spec.lookback_days,
            decay_unit_days=action.decay_unit_days
            if action.decay_unit_days is not None
            else existing_spec.decay_unit_days,
            default_weight=action.default_weight
            if action.default_weight is not None
            else existing_spec.default_weight,
            resource_weights=ResourceSlot({
                k: Decimal(str(v)) for k, v in filtered_resource_weights.items()
            }),
        )

        # 7. Save via repository
        fair_share_updater = ResourceGroupFairShareUpdaterSpec(
            fair_share_spec=TriState.update(new_spec),
        )
        updater = Updater[ScalingGroupRow](
            pk_value=action.resource_group,
            spec=ScalingGroupUpdaterSpec(fair_share=fair_share_updater),
        )
        result = await self._repository.update_scaling_group(updater)

        return UpdateFairShareSpecActionResult(scaling_group=result)



import logging
from typing import Any

import sqlalchemy as sa
from sqlalchemy.exc import NoResultFound

from ai.backend.common.events.event_types.model_serving.anycast import (
    ModelServiceStatusAnycastEvent,
    RouteCreatedAnycastEvent,
)
from ai.backend.common.types import (
    AgentId,
    ClusterMode,
    ImageAlias,
    ModelServiceStatus,
    SessionTypes,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.errors.kernel import SessionNotFound
from ai.backend.manager.models.endpoint import EndpointRow
from ai.backend.manager.models.image import ImageIdentifier, ImageRow
from ai.backend.manager.models.keypair import KeyPairRow
from ai.backend.manager.models.routing import RouteStatus, RoutingRow
from ai.backend.manager.models.session import KernelLoadingStrategy, SessionRow
from ai.backend.manager.models.user import UserRow
from ai.backend.manager.models.utils import (
    ExtendedAsyncSAEngine,
    execute_with_retry,
)
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.types import UserScope
from ai.backend.manager.utils import query_userinfo_from_session

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ModelServingEventHandler:
    def __init__(self, registry: AgentRegistry, db: ExtendedAsyncSAEngine) -> None:
        self._registry = registry
        self._db = db

    async def handle_model_service_status_update(
        self,
        _context: None,
        source: AgentId,
        event: ModelServiceStatusAnycastEvent,
    ) -> None:
        log.info("HANDLE_MODEL_SERVICE_STATUS_UPDATE (source:{}, event:{})", source, event)
        try:
            async with self._registry.db.begin_readonly_session() as db_sess:
                session = await SessionRow.get_session(
                    db_sess,
                    event.session_id,
                    allow_stale=False,
                    kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
                )
                route = await RoutingRow.get_by_session(db_sess, session.id, load_endpoint=True)
        except SessionNotFound:
            return
        except NoResultFound:
            return

        async def _update() -> None:
            async with self._db.begin_session() as db_sess:
                data: dict[str, Any] = {}
                match event.new_status:
                    case ModelServiceStatus.HEALTHY:
                        data["status"] = RouteStatus.HEALTHY
                    case ModelServiceStatus.UNHEALTHY:
                        data["status"] = RouteStatus.UNHEALTHY
                query = sa.update(RoutingRow).values(data).where(RoutingRow.id == route.id)
                await db_sess.execute(query)

        await execute_with_retry(_update)

    async def handle_route_creation(
        self,
        _context: None,
        _source: AgentId,
        event: RouteCreatedAnycastEvent,
    ) -> None:
        endpoint: EndpointRow | None = None

        try:
            async with self._db.begin_readonly_session() as db_sess:
                log.debug("Route ID: {}", event.route_id)
                route = await RoutingRow.get(db_sess, event.route_id)
                endpoint = await EndpointRow.get(
                    db_sess, route.endpoint, load_image=True, load_model=True
                )

                query = sa.select(
                    sa.join(UserRow, KeyPairRow, KeyPairRow.user == UserRow.uuid)
                ).where(UserRow.uuid == endpoint.created_user)
                created_user = (await db_sess.execute(query)).fetchone()
                if created_user is None:
                    raise ValueError(f"Created user not found for endpoint {endpoint.id}")
                if endpoint.session_owner != endpoint.created_user:
                    query = sa.select(
                        sa.join(UserRow, KeyPairRow, KeyPairRow.user == UserRow.uuid)
                    ).where(UserRow.uuid == endpoint.session_owner)
                    session_owner = (await db_sess.execute(query)).fetchone()
                    if session_owner is None:
                        raise ValueError(f"Session owner not found for endpoint {endpoint.id}")
                else:
                    session_owner = created_user

                _, group_id, resource_policy = await query_userinfo_from_session(
                    db_sess,
                    created_user.uuid,
                    created_user.access_key,
                    created_user.role,
                    created_user.domain_name,
                    None,
                    endpoint.domain,
                    endpoint.project,
                    query_on_behalf_of=session_owner.access_key,
                )

                if endpoint.image_row is None:
                    raise ValueError(f"Image not found for endpoint {endpoint.id}")
                if endpoint.model_row is None:
                    raise ValueError(f"Model not found for endpoint {endpoint.id}")
                image_row = await ImageRow.resolve(
                    db_sess,
                    [
                        ImageIdentifier(endpoint.image_row.name, endpoint.image_row.architecture),
                        ImageAlias(endpoint.image_row.name),
                    ],
                )

                environ = {**(endpoint.environ or {})}
                if "BACKEND_MODEL_NAME" not in environ:
                    environ["BACKEND_MODEL_NAME"] = endpoint.model_row.name

                await self._registry.create_session(
                    f"{endpoint.name}-{event.route_id!s}",
                    image_row.image_ref,
                    UserScope(
                        domain_name=endpoint.domain,
                        group_id=group_id,
                        user_uuid=session_owner.uuid,
                        user_role=session_owner.role,
                    ),
                    session_owner.access_key,
                    resource_policy,
                    SessionTypes.INFERENCE,
                    {
                        "mounts": [
                            endpoint.model,
                            *[m.vfid.folder_id for m in endpoint.extra_mounts],
                        ],
                        "mount_map": {
                            endpoint.model: endpoint.model_mount_destination,
                            **{
                                m.vfid.folder_id: m.kernel_path.as_posix()
                                for m in endpoint.extra_mounts
                            },
                        },
                        "mount_options": {
                            m.vfid.folder_id: {"permission": m.mount_perm}
                            for m in endpoint.extra_mounts
                        },
                        "model_definition_path": endpoint.model_definition_path,
                        "runtime_variant": endpoint.runtime_variant.value,
                        "environ": environ,
                        "scaling_group": endpoint.resource_group,
                        "resources": endpoint.resource_slots,
                        "resource_opts": endpoint.resource_opts,
                        "preopen_ports": None,
                        "agent_list": None,
                    },
                    ClusterMode(endpoint.cluster_mode),
                    endpoint.cluster_size,
                    bootstrap_script=endpoint.bootstrap_script,
                    startup_command=endpoint.startup_command,
                    tag=endpoint.tag,
                    callback_url=endpoint.callback_url,
                    enqueue_only=True,
                    route_id=route.id,
                    sudo_session_enabled=session_owner.sudo_session_enabled,
                )
        except Exception as e:
            log.exception("error while creating session:")
            error_data = {
                "type": "creation_failed",
                "errors": [
                    {
                        "src": "",
                        "name": e.__class__.__name__,
                        "repr": e.__repr__(),
                    }
                ],
            }

            async def _update() -> None:
                async with self._db.begin_session() as db_sess:
                    query = (
                        sa.update(RoutingRow)
                        .values({"status": RouteStatus.FAILED_TO_START, "error_data": error_data})
                        .where(RoutingRow.id == event.route_id)
                    )
                    await db_sess.execute(query)
                    if endpoint:
                        query = (
                            sa.update(EndpointRow)
                            .values({"retries": endpoint.retries + 1})
                            .where(EndpointRow.id == endpoint.id)
                        )
                        await db_sess.execute(query)

            await execute_with_retry(_update)



from dataclasses import dataclass
from typing import override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.services.permission_contoller.actions.base import RoleAction


@dataclass
class GetRoleAction(RoleAction):
    user_id: UUID
    role_id: UUID

    @override
    def entity_id(self) -> str | None:
        return str(self.role_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET


@dataclass
class GetRoleActionResult(BaseActionResult):
    @override
    def entity_id(self) -> str | None:
        return None



"""Common Resource Usage GQL types."""

from __future__ import annotations

from datetime import date, datetime
from enum import StrEnum

import strawberry


@strawberry.type(
    name="UsageBucketMetadata",
    description=(
        "Added in 26.1.0. Common metadata for usage bucket records including "
        "the measurement period and timestamps."
    ),
)
class UsageBucketMetadataGQL:
    """Common metadata for usage bucket records."""

    period_start: date = strawberry.field(
        description="Start date of the usage measurement period (inclusive)."
    )
    period_end: date = strawberry.field(
        description="End date of the usage measurement period (exclusive)."
    )
    decay_unit_days: int = strawberry.field(
        description="Number of days in each decay unit for this bucket."
    )
    created_at: datetime = strawberry.field(description="Timestamp when this record was created.")
    updated_at: datetime = strawberry.field(
        description="Timestamp when this record was last updated."
    )


@strawberry.enum(
    name="UsageBucketOrderField",
    description=(
        "Added in 26.1.0. Fields available for ordering usage bucket query results. "
        "PERIOD_START: Order by the start date of the usage measurement period. "
        "Use DESC to get most recent buckets first, ASC for chronological order."
    ),
)
class UsageBucketOrderField(StrEnum):
    PERIOD_START = "period_start"



"""Get domain action and result types."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.domain.types import DomainData
from ai.backend.manager.services.domain.actions.base import DomainAction


@dataclass
class GetDomainAction(DomainAction):
    """Action to get a single domain by name.

    Args:
        domain_name: Name of the domain to retrieve.

    Raises:
        DomainNotFound: If domain does not exist.
    """

    domain_name: str

    @override
    def entity_id(self) -> str | None:
        return self.domain_name

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET


@dataclass
class GetDomainActionResult(BaseActionResult):
    """Result of GetDomainAction.

    Args:
        data: Domain data retrieved from repository.
    """

    data: DomainData

    @override
    def entity_id(self) -> str | None:
        return self.data.name



"""Remove `storage_id`, and add `storage_namespace_id` in association_artifacts_storages

Revision ID: 43c7e60c56af
Revises: b6f3b3c69d16
Create Date: 2025-08-28 05:43:18.926924

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "43c7e60c56af"
down_revision = "b6f3b3c69d16"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "association_artifacts_storages", sa.Column("storage_namespace_id", GUID(), nullable=False)
    )
    op.add_column(
        "association_artifacts_storages", sa.Column("storage_type", sa.String(), nullable=False)
    )
    op.drop_column("association_artifacts_storages", "storage_id")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "association_artifacts_storages",
        sa.Column("storage_id", GUID(), autoincrement=False, nullable=False),
    )
    op.drop_column("association_artifacts_storages", "storage_type")
    op.drop_column("association_artifacts_storages", "storage_namespace_id")
    # ### end Alembic commands ###



from .resolver import agent_stats, agents_v2
from .types import (
    AgentFilterGQL,
    AgentOrderByGQL,
    AgentResourceGQL,
    AgentStatsGQL,
    AgentV2Connection,
    AgentV2Edge,
    AgentV2GQL,
)

__all__ = (
    # Types
    "AgentFilterGQL",
    "AgentOrderByGQL",
    "AgentResourceGQL",
    "AgentStatsGQL",
    "AgentV2GQL",
    "AgentV2Connection",
    "AgentV2Edge",
    # Resolvers
    "agent_stats",
    "agents_v2",
)



"""add_index_sessions_vfolder_mounts

Revision ID: 31463788c713
Revises: dddf9be580f5
Create Date: 2024-04-19 18:53:29.903113

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "31463788c713"
down_revision = "dddf9be580f5"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_index(
        "ix_sessions_vfolder_mounts",
        "sessions",
        ["vfolder_mounts"],
        unique=False,
        postgresql_using="gin",
    )


def downgrade() -> None:
    op.drop_index("ix_sessions_vfolder_mounts", table_name="sessions", postgresql_using="gin")
